#!/usr/bin/python3

import logging

""" Facilitate HDF5 simulation and generation from "plain" Python data.

Nx5d data models evolve around the HDF5 API. This module is dedicated
to exposing the result of data processing, i.e. plain Python data,
numpy-(like-)arrays etc, as h5py-like objects. The simplest way of doing
that is *actually* storing those data in true h5py HDF5 file -- be they
in memory (`h5py.File(... driver='core')`) or on-disk.

The straight-forward method of doing that is invoking `concat_h5()`
from this module, which adds a dataset, or a subtree (referenced by
a python `dict`) in a h5py file. Compatible with the nx5d *Scan/Streak/Frame*
mindset, the `concat_h5()` method concatenates new data to already existing
datasets, along the first dimension.
"""

def concat_h5(h5like, data, path="/", dset_args=None):
    ''' Concatenates the data from `data` into the HDF5-like `h5like`.

    The concatenation is done by *actually* extending an existing dataset along
    its first dimension. For this, the data to be concatenated needs to have
    matching sizes in all other dimensions (and in particular have the same
    dimensionality) as the existing dataset.

    Apparently nother way how this could've been achieved is by storing the
    incoming datasets separately and (re-)creating a virtual dataset over
    and over again.

    Args:
        h5like: a `h5py.File` like object.

        data: a set of possibly nested Python containers containing data.
          As such, `data` might be:
          - An array-like (in which case `path` needs to be something different
            than the default "/", and this is treated like a single dataset)
          - An enumerable (in which case we don't know what to do with it,
            but we feel like we should to *something*, so there's that...)
          - A dictionary / hashable, in which case this function is recursively
            called for every item, replacing `path` by the hashable key,
            and `data` by the hashable value)

        path: The path at which to insert `data` within the HDF5 object.
          If `data` is an array-like, it is concatenated to the possibly
          existing dataset at `path` along the first dimension. If it
          is a hashable, then `path` is expected to be an HDF5 group node
          (or is created as one, if it doesn't already exist) under which
          to create datasets as instructed by the hashable.

        dset_args: Keyword arguments to pass to `h5py.Group.create_dataset()`
          when creating new datasets. This is `None` by default, meaning that
          no extra parameters will be passed. But subsequently adding datasets
          -- which is the primary scope of this function -- requires HDF5
          datasets to be created as chunked, and to have `maxshape` in the
          required dimension set accordingly. For this reason, `concat_h5()`
          will automatically assume `maxshape=(None, ...)` if nothing else
          is specified for `maxshape`, which will also automatically enable
          chunking (according to h5py documentation).

    Returns: `None`

    Raises:
    
        - `TypeError` if any of the required nodes already exits in the HDF5-like
          but is not of the expected type (dataset or group), or of a compatible
          shape
    
        - Any exception which `h5py.Group.create_dataset()` or `.resize_dataset()`
          might also raise, in particular `RuntimeError` if parameters are not
          compatible, e.g. the datasets are initially created without the ability
          to be extended.
    '''

    if dset_args is None:
        dset_args = {}

    if data is None:
        # Special case: ignore datasets that are 'None'. This is mainly
        # a feature for unit testing, but it's also sensible behaviour.
        return
    
    elif hasattr(data, "shape"):
        try:
            dset = h5like[path]
            
            if not hasattr(dset, "shape"):
                raise TypeError("%s exists in HDF5-like and is expected to be a "
                                "dataset, but it isn't" % path)

            if dset.shape[1:] != data.shape[1:]:
                raise TypeError("%s exists in HDF5-like but has shape %r incompatible with %r " % \
                                (path, dset.shape, data.shape))

            new_size = dset.shape[0]+data.shape[0]
            
            logging.debug("Extending dataset '%s' in '%s' to %d" % (path, h5like.name, new_size))
            dset.resize(new_size, axis=0)
            dset[-data.shape[0]:] = data[:]
            
        except KeyError as e:
            # dataset does not yet exist -- we're creating it.
            logging.debug("Creating dataset '%s' in '%s'" % (path, h5like.name))

            # If the user didn't specify anything else, we make the dataset
            # infinitely extensible in the 1st dimension.
            max_shape = tuple([None] + [i for i in data.shape[1:]])
            dset_args.setdefault('maxshape', max_shape)
            h5like.create_dataset(path, data=data, **dset_args)

    else:
        try:
            group = h5like[path]
        except KeyError:
            # need to create group first
            split = path.rsplit('/')
            h5base = h5like if len(split) == 1 else h5like['/'.join(split[:-2])]

            logging.debug("Creating group '%s' in '%s'" % \
                          (split[-1], '/'.join(split[:-2])))
            
            group = h5base.create_group(split[-1])
        
        for k, v in data.items():
            concat_h5(group, v, path=k, dset_args=dset_args)



#!/usr/bin/python3

import h5py

from os import scandir
import re

import tifffile

def pilatus_get_tifflist(basepath, set_index=None):
    '''
    Returns a sorted list of Pilatus TIFF files (full path!)
    from the base pilatus dir `basepath` the data index `index`.
    '''

    if set_index:
        tiffdir = "%s/pilatus/S%.5d" % (basepath, set_index)
    else:
        tiffdir = basepath

    try:

        # The TIFF files generated by Pilatus usually have some format of the
        # form <mumble>_xxx.<extension>, with 'xxx' being an index with a variable
        # number of digits. <mumble> is usually the same everywhere, and <ext>
        # is usually "tif". This algorithm doesn't exacly rely on that being true,
        # but it's a good idea to check, just in case.
        #
        # In any case, we need to reorder the files by the 'xxx' sequence, which
        # means that we have to parse the filename first, sort by 'xxx', then
        # rebuild the path again.
        
        regex = re.compile(r"(.*)([^\d])(\d+).([a-zA-Z]*)")
        tuples = []
        
        for f in scandir(tiffdir):
            m = regex.match(f.path)
            tuples.append((m.group(1), m.group(2), m.group(3), m.group(4)))

        files_sorted = [ f[0]+f[1]+f[2]+"."+f[3]
            for f in sorted(tuples, key=lambda t: int(t[2]))]

        return files_sorted
        
    except FileNotFoundError as e:
        #print("pilatus_get_tifflist: path error:", e)
        raise e

    except IndexError as e:
        #print("pilatus_get_rifflist: file name parse error:", e)
        raise e    
    

if __name__ == "__main__":

    # Trying to open a Nexus HDF5 file and read some UDKM data

    hdfile  =  "/var/home/florin/Documents/sync/Uni/Analysis/221-11141-vonreppert/tmp.hdf5"
    database = "/var/home/florin/Documents/sync/Uni/DataCache/xpp-data/221-11141-vonreppert"

    # name of the TIFF data within the /Sxxx/data/ section of the Nexus file
    tiff_group_name = u'maps'
    
    f = h5py.File(hdfile, "a")
    
    for item in f.attrs.keys():
        print (item + ":", f.attrs[item])

    print ("%d items in file: %r" % (len(f.keys()), f.keys()))

    
    for key in sorted(f.keys(), key=lambda k: int(k[1:])):

        dataset = int(key[1:])

        #print ("Dataset %d" % dataset)
        
        datafile = "%s/pilatus/S%.5d" % (database, dataset)
        setpath  = "/S%d" % (dataset)
        datapath = "/S%d/data" % (dataset)

        #print ("%d / HDF5 set path: %s" % (dataset, datapath))
        #print ("%d / TIFF file path: %s" % (dataset, datafile))
        
        setobj = f[setpath]
        

        try:
            tiffs = pilatus_get_tifflist(database, dataset)

            #mapsobj = setobj.require_group(tiff_group_name)
            #mapsobj.attrs[u'NX_class'] = u'NXentry'
            #mapsobj.attrs[u'set_index'] = dataset

            setobj.attrs[u'default'] = tiff_group_name

            maps_parent = setobj

            dataobj = maps_parent.require_group(tiff_group_name)
            dataobj.attrs['NX_class'] = 'NXdata'
            dataobj.attrs['axes'] = ['x', 'y']            
            maps_parent.attrs[u'default'] = tiff_group_name

            for i in range(len(tiffs)):
                
                cnts = tifffile.imread(tiffs[i])
               
                data_name = "map-%d" % i

                dataobj.attrs['signal'] = data_name
                dataobj.attrs['title'] = tiffs[i]

                # Just a convenience -- put the last succesfully loaded data
                # map as the default map in this map
                
                dcnt = dataobj.require_dataset(name=data_name,
                                               data=cnts,
                                               shape=cnts.shape,
                                               dtype=cnts.dtype,
                                               exact=True)

                dx = dataobj.require_dataset(name='x',
                                             data=[i for i in range(cnts.shape[0])],
                                             shape=(cnts.shape[0]),
                                             dtype=int)
            
                dy = dataobj.require_dataset(name='y',
                                             data=[i for i in range(cnts.shape[1])],
                                             shape=(cnts.shape[1]),
                                             dtype=int)
                
                
            print ("%d / %d TIFF files" % (dataset, len(tiffs)))
            
        except FileNotFoundError:
            print ("%d / No TIFF data" % dataset)
            
        
        #for f in tiffs:
        #    print (f)
        #
        #    print (tiffs[5])
        #    data = tifffile.imread(tiffs[10])
        #    print (data, data.max())



#!/usr/bin/python

from nx5d.xrd import data as xd
from nx5d.xrd import kmc3 as kmc3

import pytest, random

from pprint import pprint

@pytest.fixture
def testfile():

    # Note that, for data storage reasons within the git project,
    # our testfile contains only TIFF data for scan "2.1".
    return "./tests/test_data/spech5/231-cw7-12083-roessle.spec"


def test_experimentsetup(testfile):

    h5like = kmc3.SpecTiffH5(testfile)
    
    templ = kmc3.ExperimentTemplate
    
    setupGen = xd.SetupGenerator(templ)
    setup = setupGen(h5like)

    for i in templ:
        assert i in setup

    for i in setup:
        assert i in templ

    # Test setting of values fpr keys that already exist.
    ev = random.random()*1000
    assert xd.SetupGenerator(templ, beamEnergy=ev)(h5like)['beamEnergy'] == ev

    # Test that keys not in template are rejected
    with pytest.raises(KeyError):
        xd.SetupGenerator(templ, jinkies=ev)
    
    #print("Experiment template:")
    #pprint(templ)
    #print("Experiment definition:")
    #pprint(setup)



#!/usr/bin/python3

import pytest
import h5py
from numpy import array as np_array

from nx5d.h5like.pypod import concat_h5

def test_concat_h5_groups():
    # Tests the group creating / hierarchy
    
    h5 = h5py.File('/dev/null', 'w', driver='core', backing_store=False)

    # Test adding folders
    concat_h5(h5, { "jinkies": None })
    with pytest.raises(KeyError):
        # dataset "jinkies" is None, supposed to not exist
        assert h5['jinkies']

    # Test adding datasets
    concat_h5 (h5,
               { "jinkies": { "scooby": np_array(range(17)) } })

    assert h5['jinkies']
    assert h5['jinkies/scooby']

    s = h5['jinkies/scooby']
    assert len(s.shape) == 1
    assert s.shape[0] == 17

    # Test extending datasets
    concat_h5 (h5, { "jinkies": { "scooby": 17+np_array(range(23)) } })
    s = h5['jinkies/scooby']
    
    assert len(s.shape) == 1
    assert s.shape[0] == 17+23

    for i,j in zip(s,range(40)):
        assert i == j

    # Test adding and extending multidimensional datasets
    concat_h5(h5, {"boron": {"theMoron": np_array(range(20)).reshape(4,5)}})
    concat_h5(h5, {"boron": {"theMoron": np_array(range(30)).reshape(6,5)}})

    s = h5["boron/theMoron"]

    assert s.shape == (10, 5)
            

        



#!/usr/bin/python3

import numpy as np

class RunlogFile():
    '''
    This is meant for "what is what" logfile which accompanies EDF data,
    apparently written at some ESRF facilities.

    The logfile format is essentially a TSV (tab separated values) file
    with a header line for field naming. What we're doing is exposing this
    as a "flat" HDF5-like object where each column of the logfile corresponds
    to a data container.
    '''

    def _consume_header(self, fobj):
        '''
        Returns the header `(attribues, column)` from the top of the
        file object `fobj`
        The "attributes" are `key: value` pairs at the beginning, and the
        "column" tags are the last line before the data section.
        The corresponding lines are consumed from `fobj`.
        '''

        attrs = {}
        header_candidate = None
        
        # First read the header
        while True:

            _line = fobj.readline()
            if not _line:
                raise RuntimeError("Premature end of file")
            
            line = _line.strip()
            
            if line[0] == '#':
                if len(line) == 1:
                    # skipping lines with a single '#'
                    continue
                header_candidate = line
                
            tmp = line[1:].split(':')

            if len(tmp) == 0:
                continue
            
            elif len(tmp) == 2:
                # Trying to parse tmp[1] as a number would be cool, but the
                # data crammed in there -- even when numerical -- is a mess
                # of units, information etc. Not really meant to be parsed.
                attrs[tmp[0].strip()] = tmp[1].strip()
                
            elif len(tmp) == 1:
                header = []
                for h in header_candidate[1:].split(' '):
                    if len(h) > 0:
                        header.append(h)
                return attrs, header
            
            else:
                raise RuntimeError("Confused by header line: '%s'" % line)


    def _consume_data(self, fobj, columns, dtype_map='auto'):
        '''
        Parses data, line by line, and stores it into a dictionary where
        the keys are strings from `columns`, values are 1D `numpy.ndarray`,
        objects and the entries in the ndarray are the numbers coresponding
        to each data file line.

        Parameters:
          - `fobj`: is a file object created by `open()` (NOT a file path)
          - `columns`: This is a list of string, containing the name of the data
            columns. It must be as long as there are items in each line.
          - `dtype_maps`: This is an optional column name -> numpy-dtype map for
            the data. If specified, it will be used to transform the data from
            string to whatever this map specifies. If set to None, no transformation
            will take place. In the default value of `'auto'`, this function tries
            to autodetect integer and floating point data, leaving everything
            else as strings.
        '''

        data = {}
        for c in columns:
            data[c] = [] 

        line_no = 0
        while True:
            _line = fobj.readline()
            if not _line:
                break

            line_no += 1

            line = _line.strip()

            index = 0
            for val in line.split(' '):
                if len(val):
                    try:
                        data[columns[index]].append(val)
                        index += 1
                    except IndexError as e:
                        raise RuntimeError("%s: have %d columns, but %d data fields" %
                                           (str(e), len(columns), index))
            
            if index != len(columns):
                raise RuntimeError("Too few items in line HEAD+%d(found %d, expecting %d):"
                                   " \"%s\"" % (line_no, index, len(columns), line))

        # At this point we have a dictionary of data lists full with strings.
        # We'd rather have "proper" data, i.e. floats, dates, timestamps etc.
        for col in columns:
            l = data[col]
            a = np.array(l)
            final_a = None

            if isinstance(dtype_map, map):
                try:
                    final_a = np.array(a, dtype=dtype_map[col])
                except:
                    pass
            elif dtype_map == 'auto':
                try:
                    final_a = np.array(a, dtype=int)
                except:
                    try:
                        final_a = np.array(a, dtype=float)
                    except:
                        final_a = a
            
            data[col] = final_a if final_a is not None else a

        return data
        

    def __init__(self, f, mode='r', dtype_map='auto'):
        '''
        Opens the Logfile specified by `f` with specified `mode`
        (currently only opening for reading, `mode='r'`, is accepted).
        The optional parameter `dtype_map` specifies how to interpret
        data columns in a column name -> numpy dtype map. If set to
        `None`, data is left as string. If left on its default value
        (`auto`), interpretation as integer is first attempted, followed
        by floating point arrays. Fallback is string.
        '''

        # This constructs a mostly empty / dysfunctionsl object.
        # Used for unit testing.
        if not f:
            return
        
        if mode != 'r':
            raise RuntimeError("EdfLogFile: only reading supported (requested: %s)" % mode)

        if not hasattr(f, "readline"):
            f = open(f)

        self.__attrs, self.__columns = self.__consume_header(f)
        self.__data                  = self.__consume_data(f, self.__columns)


    def __getitem__(self, column):
        '''
        Returns the data array of the specified column. The column may be
        specified as "name" or "/name", to mimic HDF5 path specifications.
        '''
        if column[0] == '/':
            col = column[1:]
        else:
            col = column

        return self.__data[col]


    def __setitem__(self, column, data):
        '''
        Sets the column data at once. No checks for data type or length are performed.
        '''
        if column[0] == '/':
            col = column[1:]
        else:
            col = column
            
        self.__data[col] = data


    def keys(self):
        '''
        Returns a list with all column names.
        '''
        return self.__columns

    @property
    def attrs(self):
        '''
        Returns the list of key: value pairs at the top of the index file (as a map)
        '''
        return self.__attrs

    @attrs.setter
    def attrs(self, a):
        self.__attrs = a
            

#
# Testing only below this point
#
import pytest
from io import StringIO

def test_header_standard():
    '''
    Tests a simple key: value header setup, no surprises
    '''
    


#
# For more convenience during development. This may or may not fit 
#
from sys import argv, exit

if __name__ == "__main__":

    try:
        if len(argv) < 2:
            print("Test usage: %s file.log" % argv[0])
            exit(-1)

        print ("From logfile: %s" % argv[1])
        l = RunlogFile(argv[1])

        for k in l.keys():
            print ("Column: %s, %d items, %r" % (k, len(l[k]), l[k].dtype))
                   
    except RuntimeError as e:
        print (e)
        raise



#!/usr/bin/python3

import time
from numpy import array as np_array

import logging

class HdfPathTemplateRestrictor:
    '''
    Takes a dictionary which may contain HDF5 paths in the values
    and "restricts" the template by replacing the templated values
    by the data at the corresponding HDF5 path.

    Works recursively on the data, i.e. if the dicitonary contains
    sub-structured data, the structure is retained. All values that
    don't contain HDF5 references are returned verbatim.

    For instance, using this template:
    ```
    {
       "h5data": "@/path/to/dataset",
       "static": 3.14,
       "nested": {
          "static": [3, 4, 5],
          "h5data": "@/path/to/another/dataset",
       }
    }
    ```

    Would return something like this (assuming that "/path/to/dataset"
    and "/path/to/another/dataset" are arrays containing `[1, 2]` and `[3, 4]`,
    respectively):
    ```
    {
       "h5data": [1, 2],
       "static": 3.14,
       "nested": {
          "static": [3, 4, 5],
          "h5data": [3, 4],
       }
    }
    ```
    
    '''
    
    def __init__(self, template,
                 accept=None,
                 paths=None,
                 passthrough=False):
        '''
        Initialises the template restrictor.

        Parmeters:
        
          - `template`: the configuration template dictionaray

          - `accept`: Places a whitelist restriction on the template
             keys to process and return. This is useful e.g. when
             template dictionaries are to be processed
             only partially. Note that unprocessed keys are _not_
             returned. Default is `None`, which disables the whitelist
             restriction.

          - `paths`: If specified, it's expected to contain a
            for path format substitutes, i.e. this enables the template
            to contain references liek "h5:#{key}", where `{key}` will
            be replaced with whatever is in `paths` at the corresponding
            dict slot.

          - `passthrough`: If set to `True`, instead of returning the
            the actual data for HDF5 references, the HDF5-like Python
            container will be returned (e.g. the dataset HDF5 node object).
            This is useful for accessing metadata provided by HDF5 (e.g.
            dataset sizes and shapes) without actually reading out
            possibly large amounts of data. Default is `False`.
        '''
        
        self._tmpl = { k:v for k,v in template.items() }
        self._paths = {k:v for k,v in (paths or {}).items() }
        self._accepted = accept or [k for k in template.keys()]
        self._pass = passthrough


    def __value(self, h5like, value, fmtKeys, nodeOnly=False, slicer=slice(None)):
        '''
        This is the heard of the HDF5 address data lookup system: it checks
        the contents of `value` and returns the corresponding data as expected.
        The trick is in correctly determining the meaning of `value` here.

        The simples assumption here is that `value` is, indeed, a real piece
        of data (i.e. a number or an array with numbers), in which case it
        should be returned verbatim.

        But mostly this isn't the case. `value` might typically differ from
        a pure value in two ways (or any combination thereof):

          1. `value` is a string reference to a path within a HDF5-like
             object (file or node) -- here referenced by `h5like`.

          2. `value` is a container (tuple or dict) that we need to recursively
             descend into and look at each item individually.

        
        For the case of (1), the distinction is the following:
        
          - If `value` is a string that begins with `@`, then it is taken
            as a HDF5 address relative to `h5like`. The string is assumed
            to be a format (e.g. "contain {key}") where the keys to be
            substituted are taken from `fmtKeys`.

          - If `value` is a tuple which's first item is a string that begins
            with "@", and *all* other items are `slice()` objects, then
            the first item of the tuple is assumed to be the HDF5 path
            (as above), but instead of returning the value as-is, it is
            sliced using the provided slicer.

          - Everything else is returned as-is.


        We enter the case of (2) if neither of the above is the case, i.e.
        if `value` is either a dict, or another type of container (tuple
        with different items than (string, slice, ...),

          - If `value` is a tuple that doesn't fit the ("addr", slice()...)
            form, or is a dictionary, we need to recursively call
            `__valyue()` for each of the items.

        
        The function finally returns something that has the same
        structure as `value`, but eventual HDF5 paths are replaced by
        actual data as requested.
        '''

        my_slicer = slicer

        # Check if value is a fancy ("addr", slice(), ...) tuple.
        # If it is, rewrite `value` to actually contain only the address,
        # and append the slicing part `value` to the slicer argument.
        if isinstance(value, tuple) and \
           len(value)>=2 and \
           (isinstance(value[0], str) and value[0][0] == '@') and \
           np_array([isinstance(v, slice) for v in value[1:]]).all():
            if isinstance(slicer, tuple):
                my_slicer = (*slicer, *value[1:])
            else:
                my_slicer = (slicer, *value[1:])
            value = value[0]


        if isinstance(value, str):
            # The main feature: loading data from HDF5 if we encounter an "@..." string.
            # FIXME: need to actually rewrite this and base this on the __call__() feature
            # (see below). This will simplify many things.

            if len(value) == 0 or value[0] != '@':
                return value
            
            nodeAddrFmt = value[1:]
            
            try:
                nodeAddr = nodeAddrFmt.format(**fmtKeys)

                node = h5like[nodeAddr]
                
                logging.debug("Loading data from %s (only node: %r)" % (nodeAddr, nodeOnly))
                
                if nodeOnly:
                    return node
                else:
                    try:
                        s = tuple(my_slicer[:len(node.shape)]) if isinstance(my_slicer, tuple) else my_slicer
                        return node[s]
                    #except TypeError:  # ...not sure when this happens (?)
                    #    return node
                    except ValueError: # This happens when value is a scalar and needs to be sliced by ()
                        return node[()]
                        

            except KeyError as e:
                logging.error("Path substitition error in '%s': %s" % (nodeAddrFmt, str(e)))
                raise
                    
            except Exception as e:
                logging.error("%s: error reading node (%r)" % (nodeAddr, str(e)))
                raise
                
        elif isinstance(value, dict):
            # Descending into each of the elements
            return dict({k:self.__value(h5like, v, fmtKeys, nodeOnly, slicer) for k,v in value.items()})
        
        elif hasattr(value, "__len__"):
            # It's not a string and it's not a dict, but is iterable?
            # descend and find out.
            return tuple([self.__value(h5like, v, fmtKeys, nodeOnly, slicer) for v in value])

        elif hasattr(value, "__call__"):
            # This allows us to create custom objects that retrieve
            # data from unconventional sources.
            return value(h5like, nodeOnly, slicer)

        else:
            return value

        
    def __call__(self, h5like=None, paths=None, nodeOnly=False, slicer=slice(None)):
        '''
        Uses the stored template to generate meaningful content based on the
        data from `h5like`. Typically, this involves going through all the
        templates' values and checking whether they consist of real data
        or H5 paths that need to be read. All paths are relative to `h5like`.

        If `nodeOnly` is set to `True`, the HDF5 nodes at the "nx5:" dataset
        paths is returned. Otherwise the `.value` property is read.

          - `slicer`: If specified, it is applied to any data that is to
            be retrieved from the HDF5-like node, insofar as the data shape
            matches the dimensionality of the slicer; if the slicer has
            too many components (e.g. a 3D-slicer on 1D data), the slicer
            is sliently truncated.        
        '''

        if h5like is None:
            return {k:v for k,v in self._tmpl.items()}

        combinedPaths = {k:v for k,v in self._paths.items()}
        combinedPaths.update((paths or {}))

        retr = {}
        for k,v in self._tmpl.items():
            if k in self._accepted:
                val = self.__value(h5like, v, fmtKeys=combinedPaths,
                                   nodeOnly=nodeOnly, slicer=slicer)
                retr[k] = val
                continue

            if self._pass:
                retr[k] = v

        return retr


class SetupGenerator(HdfPathTemplateRestrictor):
    '''
    Generic builder to generate an "experiment setup" dictionary.

    Concepts
    ========
    
    The structure of an experiment setup contains several types
    of data:

      - "Static" persistent device geometry data, e.g. the type
        and axes of the goniometer etc

      - "Static" transient data which is valid for a particular
        data set or scan, but not necessarily for others, e.g.
        current center of the detector plate, photon energy etc

      - "Dynamic" data which might changes from frame to frame
        (mostly positioners and auxiliary experimental data like
        ring gurrent etc.

    "Static" persistent data
    ------------------------
    
    All notations and conventions are analogous to `xrayutilities`
    device / goniometer setup:

      - `goniometerAxes`: The direction of each of the goniometer axes in
        the `[xyz][+-]` notation. This is a variable-sized array, as there
        can be several axes in any goniometer, and `xrayutilities` apparently
        magically knows what to do.

      - `detectorTARAxes`: The direction of each of the 3 possible movement
        axes of the detector (tilt, azimuth, rotation). Note that there are
        always 3 of these axes each with a very specific purpose in
        `xrayutilties`. If your detector lacks any, complement with `None`.

      - `imageAxes`: The direction of the image axes (x and y) at zero angles.
        The positive direction of the axes should coincide with increasing pixel
        index in the data.

      - `imageCenter`: This is the positio of the center pixel, either absolute
        (integer pixel numbers), or relative to the sensor size (as specified in
        `imageAxes`). If the number is in the range 0.0..1.0, then relative
        positioning is assumed.

      - `imageChannelSpan` / `imageDistAndPixsize`: for Q transformation,
        ultimately the relation between every specific on the detectors and the angle
        of the incoming beam activating that specific pixel is needed. There
        are two distinct ways of specifying this: either using the "channel span",
        i.e. the size, in degrees, of each pixel, in horizontal/vertical direction,
        or by a distance parameter (from detector to sample) and a pixel size.
        `imageChannelSpan` is either a single number or a 2-tuple specifying
        how many degrees one channel takes. `imageChannelSize` specifies the
        spatial size of a pixel relative to the distance between the sample
        and the sensor.

      - `sampleFaceUp`: Direction of the "sample surface facing up", a.k.a.
        "sampleor" in `xrayutilities` lingo. This is the orientation of
        the sample surface at zero angles. This is either an axis notation
        (`[xyz][+-]`) or one of the special words `det`, respectively `sam`.


    Example for device geometry:

    ```
    {
      "goniometerAxes": ('y+', 'z+', 'x+'),
    
      "detectorTARAxes": (None, "z+", None),
      "imageAxes": ("x+", "z+"),
      "imageSize": (1024, 768),
      "imageCenter": (0.5, 0.5),
      "imageChannelSpan": (None, None),
      "imageChannelSize": (None, None),
    
      "sampleFaceUp": 'x+',
      "beamDirection": (0, 1, 0),
    }

    ```

    "Static" itinerant data
    -----------------------

    Experiment setup contains additional parameters specific to the current
    experiment (as opposed to: typical for the experimentl setup). These
    include:

      - `sampleNormal`: Direction perpendicular to the surface. Please also
        consult the official `xrayutilites` documentation as to the exact meaning
        of the directions.

      - `beamEnergy`: energy of the x-ray beam (in eV).

      - `detectorTARAngles`: (optional) angle values for tilt, azimuth and
        rotation of the detector. Default values for each of these angles is
        0. Only angles which are not `None` in `detectorTARAxes` are accepted.

    Example for the sample geometry setup:
    ```
    {
      "sampleNormal": (0, 0, 1),
      "beamEnergy": 15000.0,
      "detectorTARAngles": (0, 0, 0),
    }
    ```

    
    "Dynamic" measurement data
    --------------------------

    These include mostly goniometer axes, homing offsets etc. There is either
    one of these per scan, or one per frame (i.e. a whole container/array of
    such elements per scan). These include:

      - `goniometerAngles`: positioning for each of the angles of the
        goniometer, as specified in `goniometerAxes`

      - `detectorTARAngles`: positioning for the detector tilt, azimuth
        and rotation (as in `detectorTARAxes`)

    Some of these configuration parameters (e.g. the goniometer angles) are
    inherently "not fixed" as they are an essential part of the measurement.
    Others, like the detector angles, may or may not be fixed (e.g. detector
    might be permanently fixed at an azimuth different from 0, or may
    change with every frame or with every scan, etc).

    For this reason, not only scalar numbers are accepted, but actually
    HDF5 dataset addresses as string. Here is an example of how such a
    setup definition might look like:

    ```
    {
      "goniometerAngles": ("Theta", "Phi", "Chi"),
      "detectorTARAngles": (0, "Psi", 0)
    }
    ```

    Here, the dataset "Theta" with respect to the default positioner path
    would be used as the first goniometer angle (for ESRF-style HDF5, this
    would be "instrument/positioners/Theta".)

    There might also be other positioning parameters relevant for data
    transformation and analysis, for instance "TwoTheta" instead of "Psi",
    which would have a similar meaning except for a different
    direction / offset, with respect to "Psi." In essence, it's the
    responsibility of higher layers to decide what addresses / angles should
    be used here.

    There is no data pendant to the `imageAxes` parameter because the angle
    information for each image is within the index of its own pixels.
    The data set may also contain other data of interest (e.g. "SBcurrent"),
    but that is viewed as something to be accessed in parallel to, and not
    as a basis of, the main image data.

    Each of the data fielts itself (e.g. "image", but also others like
    "SBcurrent", "delay", ...) which may be imporant for the physical analysis,
    but not to the transformation of data, is not a matter of `ExperimentalSetup`
    (FIXME: or is it?) but one of data collection fields within
    `ScanReader`.

    Placeholder variables
    ---------------------

    When using HDF5 paths as placeholders, `SetupGenerator` can use string
    formatting if a dictionary of keys is supplied at `__init__`
    or at `__call__`. `ScanReader` supplies the following keys, compatible
    with the ESRF flavor of HDF5/Nexus data format:

      - "instrument": The top level directory containing instrument information,
        typically of HDF5 type "NXinstrument" and located at "./instrument"
        within the scan.

      - "positioners": Typically at "./instrument/positioners", usually
        important for all the angle definitions

      - "measurement": Top-level folder (of type "NXcollection") containing
        the measurement data (typically "./measurement").

    Usage
    =====

    The idea is to make a 3-step process:

      1. Create dictionary (with keys above) that specify static
         parameters as far as possible, and contains HDF5 paths/addresses
         for the rest

      2. Pass on dictionary to `ExperimentSetup` class, which stores the
         explicitly passed values, and also offers an interface (upon
         construction or later) to override or modify keys

      3. Override `ExperimentSetup.__call__()` to export a "final" version
         of the data, hand-matched for the current scan dataset (passed on
         as parameter to `__call__()`) which does not contain anything
         else besides pure, usable values :-)

    (FIXME: the whole "\\*Angles" section is actually supposd to *not* know
    about loading data yet... isn't it? In a way, the Angles section is
    less like "experimental setup" and more like "measurement results.")
    '''

    # ( FIXME: not sure whether there is "the" / "only one" main image data,
    # or if there should be allowed for several. On one hand, scalar arrays
    # like "SBcurrent" suggest that there's nothing special about "the" image;
    # on the other hand, angles and goniometer parameters are valid for *one*
    # specific detector location, so if we were to accept several detector(-like)
    # images as data, we'd need to duplicate some of the "static" information,
    # and also account for that within the `ScanReader` object.
    # As it stands, the architecture is only for *one* detector, obviously.)

    def __init__(self, *args, **kwargs):
        super().__init__(*args,
                         accept=[ "beamDirection",
                                  "beamEnergy",
                                  "goniometerAxes",
                                  "detectorTARAxes",
                                  "imageAxes",
                                  "imageCenter",
                                  "imageChannelSize",
                                  "imageChannelSpan",
                                  "imageSize",
                                  "sampleFaceUp",
                                  "sampleNormal" ],
                         **kwargs)


class PositionerGenerator(HdfPathTemplateRestrictor):
    '''
    Template restrictor for positioner data (detector and goniometer angle data).
    Affects only "goniometerAngles" and "detectorTARAngles" keys.
    See also `SetupGenerator` for a detailed discussion.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args,
                         accept=[ "detectorTARAngles",
                                  "goniometerAngles" ],
                         **kwargs)

    
class ScanReader:
    '''
    Loads data from one X-ray diffraction (XRD) scan run and offers
    convesion of images to Q space for further analysis.
    
    The idea is to open/initialize metadata (experimental parameters,
    ROIs, sizes, ...) on construction, and then subsequently "pull out"
    image data. This is supposed to work on a "live" file, i.e. one
    being measured and having data added to it.

    It assumes that data is available in a HDF5-like interface.
    It opens/closes the data object with every use to account for
    parallel use by data acquistion and/or other consumers.

    The assumptions to the data are, essentially (see also `nx5d.xrd.kmc3`):

      - All relevant datasets (frame images, angles etc) are organized
        as (arrays) of datasets with the first coordinate corresponding
        to the progression of frames, i.e. Nx(WxH) for images N for a
        particular angle etc.

      - The frame number progression (index starting with 0) corresponds
        to one or two experimental scan parameters, i.e. time delay (for
        pump-probe experiments), or a particular tilt angle. The `Chunk`
        object reflects this.

      - Instrument definition and angles are compatible with `xrayutilities`,
        the package used for the Q-space conversion.

      - The data has a Nexus-like / ESRF-style layout; this means that
        all relevant measurement data is under a specific folder of type
        "NXcollection" (typically called "measurement", but this is
        configurable), and all relevant parameterisation is stored under
        another folder (typically "instrument/positioners", but again,
        configurable). Data specification is expected relative to these
        folders.

      - Quite generally, the assumption is that we're interested in one or
        more measurement sets, and each of those depends on a subset of
        positioners.
    
    '''
    
    def __init__(self, H5pyLike, scan,
                 h5kwargs=None,
                 pathKeys=None,
                 **dreq):
        '''
        Initializes the object and basic experimental parameters.
        Needs the following:

          - `H5pyLike`: A class or a factory which generates a `h5py`-like object for
            data access. Defaults to `h5py.File`, but objects with a similar API
            (e.g. the `silx.io.spech5.SpecH5` kind of loaders) are also accepted.        

          - `scan`: A parameter list that contains information to be passed to
            `H5pyLike` in order to locate the scan within the data container.
            If the first item in that list is a string (i.e. a file path), then
            the format "resource[#path]" is accepted to pass on both the file name
            ("resource") and the path inside the container ("path"). All other
            items are passed on as unnamed arguments to `H5pyFile`.
            A typical example of this is for instance
            `("/path/to/file.h5#scan23", "r")` if `H5pyLike` is a `h5py.File`
            object. For convenience, if `H5pyLike` is indeed a `h5py.File`.

          - `h5kwargs`: `None` by default, but can be passed a dictionary of named options
            to be passed on to `loader` in addition to the resource locator string.

          - `pathKeys`: Dictionary with shortcuts of standard (?) paths. Currently,
            `ScanReader` recognizes and uses the following paths:
        
              - "measurement": Where the data is

              - "positioners": Where the positioners are (e.g. angles)

          - `dreq`: This is a list of "data request" named parameters, i.e. each key is
            a measureables, to which the value is a list of other measureables or
            poisitioners. The names of the data are relative to `measurablePath`,
            respectively `positionerPath`. The data labels may also contain a path
            component (e.g. "detector/image"), in which case all path separators
            ("/") will be replaced by underscore when data naming is performed
            ("detector_image").

            FIXME: not exactly sure how this is supposed to work. There are two main
            problems with it.
        
             1. The angles: designation of axes (goniometer, detector and image)
                are already in the 'setup' parameter. But the _actual_ angles themselves
                are not. So we have two options here: squeeze them in 'setup' (in which
                case: what is this parameter for?), or squeeze them here (in which case:
                in which order? See also 2...)

             2. The data itself: most of the time the data is the actual detector
                image / frame data. But sometimes we have data of secondary interest,
                e.g. SBcurrent (ring curren) or similar, e.g. for normalization. *This*
                data does not depend on any kind of angles or positioners... or may
                only depend on _some_ positioners. How do we model that?
        '''

        self.pathKeys = pathKeys or {'measurement': 'measurement',
                                     'positioners': 'instrument/positioners',
                                     'instrument': 'instrument' }
        
        self.h5kwargs = h5kwargs or {}

        self.dataSource = self.__source_factory(scan, H5pyLike)

        self.setupTemplate = dreq.get("setup", {})

        self.requiredDatasets = dreq.copy()

        
    def __source_factory(self, params, H5pyLike):
        # Implements magic to parse out the '#path' part of the filename.
        # Returns a callable that creates a usable h5py-like object
        # for querying scan data.

        self.H5pyLike = H5pyLike
        fileUrl, fileArgs = (None, tuple())

        if isinstance(params, str):
            fileUrl = params
        elif isinstance(params, tuple):
            fileUrl, fileArgs = params[0], params[1:]
        else:
            fileUrl = params

        if isinstance(fileUrl, str):
            filePath, self.h5scan = fileUrl.split('#')
            self.h5args = tuple([filePath] + list(fileArgs))
        else:
            self.h5scan = None
            self.h5args = (fileUrl,)

        print("H5like: %r (%r, %r)" % (self.H5pyLike, self.h5args, self.h5kwargs))

        if self.h5scan:
            return lambda: self.H5pyLike(*(self.h5args), **(self.h5kwargs))[self.h5scan]
        else:
            return lambda: self.H5pyLike(*(self.h5args), **(self.h5kwargs))


    def __defPath(self, locator):
        ''' Translates "syntactic sugar" when specifying certain HDF5 data locators.
        
        Takes a HDF5 locator as accepted by `HdfPathTemplateRestrictor` and if it's
        a string, or a tuple (str, slice, ...) adds default paths in front
        of the string if necessary (e.g. turns "data" into "@{measurement}/data").

        Everything else is returned unchanged.

        Args:
            locator: eiter a string containing a HDF5 address (`"@..."`), or a tuple
              `("...", slice(), ...)` containing a HDF5 address as its first element
              and only slice objects for every other element.

        Returns:
            The locator, translated to contain a proper HDF5 addressing ("@..."),
            or verbatim if it's not a string or a slicing tuple.
        '''

        path_mangler = lambda p: p if p.find('/')>=0 \
            else '@%s/%s' % (self.pathKeys['measurement'], p)
        
        if isinstance(locator, str):
            return path_mangler(locator)
        
        elif isinstance(locator, tuple) and isinstance(locator[0], str):
            return (path_mangler(locator[0]), *(locator[1:]))

        else:
            return locator


    def read (self, frameSlicer, h5like=None, lean=True, **dataKeys):
        '''
        Returns a bunch of image(s) starting with `start`
        (default: last image taken). If `number` is None,
        all the images to the end of the scan are read.
        Parameters:
        
          - `frameSlicer`: A slice object for selecting frames within a scan
            series.

          - `roiSlicer`: Additional slicer or tuple of slicers to apply to the
            a data frame, e.g. for selecting only specific parts of the data
            frame (region-of-interest). Note that this slicer will be applied
            to *all* data keys to retrieve (see also the `dataKeys` parameter
            below), to the extent to which they have the correct dimensionality.

          - `h5like`: A h5py-like object to use for reading. If none is
            specified, the one generated by the internal `dataSource()`
            factory is used.

          - `lean`: If `True`, the "required datasets" (i.e. those that were
            specified with `__init__()`) are omitted from the result. This is
            the default.
        
          - `**dataKeys`: Data sets to retrieve; the name of the parameter will
            result in a key, and the value of the parameter should refer to the
            corresponding HDF5-like dataset path. This is passed to
            `HdfPathTemplateRestrictor`, so please also refer to the documentation
            of that class.
        
        '''        
        data = {} if lean else self.requiredDatasets.copy()
        data.update({k:self.__defPath(v) for k,v in dataKeys.items()})

        with h5like or self.dataSource() as h5:
            return HdfPathTemplateRestrictor(data, paths=self.pathKeys)\
                (h5, slicer=frameSlicer)
    

    def streaks(self, delta=None, length=None, by=None, h5like=None, retry=0, **dataKeys):
        '''
        Uses `read()` to return chunks of the data as specified by
        the `offset` and `increment parameters.
        The `addrproc` and `sliceproc` parameters are passed on to `read()`.
        
        Parameters:
        
          - `delta`: Take every `delta`-th frame to build a chunk
        
          - `length`: Take `length` consecutive frames to build a chunk
        
          - `h5like`: h5py-like object to use for reading. If none is specified
            (default), then the internal `dataSource()` method is being used,
            which initializes a new h5py object as specified in `__init__()`.

          - `worker`: Worker object ("Chunk object") to return. This
        
          - `retry`: If non-zero or not `None`, file opening/reading
            is retried on `OSError` type of exceptions up to `retry`
            number of seconds (fractional number also possible)
          
        Typically only either `delta` or `length` is necessary.
        '''

        def slicer_generator(numFrames):
            if length is None and delta is not None:
                for start in range(0, delta):
                    slicer = slice(start, None, delta)
                    yield slicer
            elif length is not None and delta is None:
                if numFrames == -1:
                    raise RuntimeError("Can't determine the number of frames from "
                                       "requested data keys")
                for start in range(0, numFrames, length):
                    slicer = slice(start, start+length, 1)
                    yield slicer
            elif length is None and delta is None:
                yield slice(None)
            else:
                raise RuntimeError("Specify either `length' or `delta' but not both.")


        with h5like or self.dataSource() as h5:

            numFrm = -1

            # read an explicit number of frames if there are any data keys,
            # otherwise we'll stick with -1 (undefined) and slicer_generator()
            # might still be able to get something out of it if enough
            # other parameters have been specified (specifically, 'delta')
            if len(dataKeys) > 0:
                tmp = next(iter(dataKeys.items()))[1]
                tmp2 = tmp if isinstance(tmp, str) else tmp[0]
                numFrm = self.numFrames(tmp2, h5like=h5like)

            scan = h5
            
            for slicer in slicer_generator(numFrm):
                begin = time.time()
                while True:
                    now = time.time()
                    try:
                        yield self.read(slicer, **dataKeys, h5like=h5, lean=False)
                        break

                    except OSError:
                        if not retry or (now-begin) > retry:
                            logging.error("Final error, %f seconds after initial start" % (now-begin))
                            raise
                        else:
                            logging.debug("OSError, retrying for another %f seconds" % (now-begin-retry))

    
    def numFrames(self, dataset, h5like=None):
        '''
        Reads (returns) the number of frames from the HDF5 file
        for the dataset refered to by `dataset`.
        '''

        with h5like or self.dataSource() as h5:

            # FIXME
            # The only reason we use "node only" here is to avoid loading
            # potentially large datasets (.shape is also available in a
            # HDF5 node without actually loading the data into memory).
            #
            # But this fucks up the HdfPathTemplateRestrictor design...
            # so we somehow need to get rid of it.
            
            data = HdfPathTemplateRestrictor({'data': self.__defPath(dataset)},
                                             paths=self.pathKeys)(h5, nodeOnly=True)
            return data['data'].shape[0]



#!/usr/bin/python3

import extra_data as ex
import nx5d.h5like.xfel as xfel

import pytest

# Unit test for the XFEL data module (mostly nx5d.h5like.xfel).
# Needs XFEL data and API, predominantly extra_data.
# Tests therefore only work within the XFEL environment.
#
# Call with 'pytest -m xfel'

@pytest.fixture
def xfel_test_run():
    return { 'proposal': 3491,
             'run': 149,
             'data': 'all' } # proposal, run

@pytest.fixture
def xfel_test_data(xfel_test_run):
    run = ex.open_run(**xfel_test_run)

    return run.select([
        ('MID_DET_AGIPD1M-1/DET/*CH0:xtdf', '*'),
        ('SA2_XTD1_XGM/XGM/DOOCS:output', '*'),
        ('MID_EXP_EPIX-1/DET/RECEIVER:daqOutput','*'),
        ('MID_EXP_SAM/MDL/DOWNSAMPLER', '*')
    ])
    
@pytest.mark.xfel
def test_exdat_h5(xfel_test_data):
    eh = xfel.ExdatH5(xfel_test_data)

    assert "measurement" in eh

    node = eh["measurement/mid_det_agipd1m-1/det/1ch0:xtdf"]
    print (node.keys())

    assert "image.data" in node
    
    #assert "instrument" in eh
    #assert "positioners" in eh["instrument"]

    data = node['image.data']
    print(data.shape)
    
    subdata = data[13:15]
    print(subdata.shape)    

    
    assert len(data.shape) == 3
    assert len(subdata.shape) == 3

    # This is just to check that the data type supports some rudimentary
    # kind of ndarray API, i.e. can do simple arithmetics and return numbers :-)
    nr = subdata.sum()
    nr2 = (subdata*2).sum()

    assert (nr*1.9 < nr2) and (nr*2.1 > nr2)



#!/usr/bin/python3

from scipy import ndimage
import numpy as np
from xarray import DataArray
import pytest

'''
Helpers for X-ray diffraction data analysis
'''

#def calcStd(x, y, center):
#    '''
#    Calculates the "standard width" (sometimes called standard deviation)
#    of an experimental data array with respect to the value located in
#    the same array at index position `center`.
#    
#    Parameters:
#    
#      - `data`: an N-dimensional `numpy` array, `xarray` or `pandas`
#        `DataFrame` or `Series`.
#        
#      - `center`: an N-tuple capable of indexing `data`.
#      
#    Returns: ...?
#    '''
#    
#    ysum = y.sum()
#    sum = 0
#    for i, value in enumerate(x):
#        sum += (value-center)**2 * (y[i]/ysum)
#
#    # FIXME: This doesn't make any f-ing sense, why would 
#
#    return np.sqrt(sum)


#def calcMoments(data, index=None):
#    '''
#    Calculates the area (i.e. moment of intertia),
#    the center of mass (in axis units),
#    and the standard deviation of the moment.
#    
#    - `data`: `numpy` array or `xarray` object with intensity data.
#      If this is an `xarray` data and `index` is not specified,
#      the index (i.e. scaling) data is obtained directly from the
#      
#    - `index`: Intrinsic scaling of `rawdata` (i.e. "axis information")
#    
#    Returns a tuple with 3 values `(area, center, std)`.
#    '''
#    
#    area = data.sum()
#    cmi  = ndimage.measurements.center_of_mass(data)
#    
#    cmi_lower  = np.array(cmi, dtype=int)
#    
#    lower_index = np.array([i for i in ])
#    
#    cmi_higher = cmi_lower+1
#    
#    #axis_offset  = roi[0]
#    #axis_delta   = ((roi[1]-roi[0])/len(data))
#    com    = axis_offset + axis_delta*ndimage.measurements.center_of_mass(data)[0]
#    std    = calcStd(index,data,com)
#        
#    return area, com, std


def series_cntrmass(data, series=None):
    '''
    Calculates N-1 dimensional center of mass for an N-dimensional array
    `data`, along the axis `series`.

    Parameters:
    
      - data:  This is expected to be an `xarray` with named dimensions.
               (FIXME: or an numpy array? Or a regular iterable, e.g. a list?)

      - series: If `data` is an `xarray`, this is expected to be a string
                naming the axis along which to break data apart.
                If data is a `numpy` array or a regular list, `series` is
                expected to be an axis index.
                If data is an iterable (e.g. a list of arrays), then `series`
                must remain `None`.
                In all cases, if `series=None` (the default), the center of
                mass series will be built along the 1st dimension.
    
    Returns an array of the same size as length of `series` dimension, each
    entry being an N-1 dimensional coordinate of the center of mass in 
    the respective plane.
    If the input is a `xarray`, then the output is also a (1D) `xarray`.
    '''
    
    
    # The difficult part is going to be that ndimage's center_of_mass()
    # returns a fractional (!) positional (!!) index, i.e. one that's
    # related to the position.
    
    # Push the dimension we want to keep to the begging (that's
    # apparently the only one we can iterate over)
    if series is not None:
        view = data.transpose(series, ...)
    else:
        view = data
        
    index_list = []
    other_axes = [o for o in view.dims[1:]] #[o for o in filter(lambda k: k != axis, view.dims)]
    
    ## iterate along `axis` and collect all the COMs in index_list
    for img in view:
        com = ndimage.measurements.center_of_mass(img.data)
        ci  = np.array(com, dtype=int)
        cf  = com-ci
        
        # iterate over each COM coordinate of the N-1 dim 
        index = [ view.coords[c].values[i] + f*(img.coords[c].values[i+1]-img.coords[c].values[i])
                     for c,i,f in zip(other_axes, ci, cf) ]
        
        index_list.append(index)
        
        
    return DataArray(data=index_list,
                     coords=(view.coords[series].values, other_axes),
                     dims=(series, 'axis'))


def sqdistance(*axes, center, shape=None):
    '''
    Takes a list of axes and a location desginated `center` and returns
    a N-dimensional array with all the squared distances from `center`
    to each of the points. The array dimensions are assumed to be
    equivalent in size.
    
    Parameters:
    
      - `*axes`: Axis objects, one per parameter. One axis object
        can either be an `xarray` coordinates tuple `(dim, values)`
        with `dim` as a string identifyer and `values` as the axis
        values; or they can be plain iterables (regular `numpy` 1D
        arrays or lists).
        
      - `center`: needs to be 1-dimensional vector with N values,
        if `axes` items are regular `numpy` arrays or iterables.
        Otherwise 
        It represents the point relative to which the squared distance
        of all other points is calculated.
        Despite its name it does not need to be, and typically is
        not, the actual center of the field.
        
      - `shape`: Instead of explicitly specifying a list of axes,
        a single vector continaing an N-dimensional array shape
        can also be specified. In that case, the axes will be
        constructed as regular indices ranging from `0` to `len(shape[n])`
        
    Returns: N-dimensional array of size `len(axes[0]) * len(axes[1]) * ... * len(axes[n])`
    contraining the squared distance to `center` for each point.
    '''
        
    if (not axes or len(axes) == 0) and shape is not None:
        # numpy array style shape. We don't actually have axis
        # values, we need to build them first. `center` is a
        # 1D-array (one coordinate per axis) we can juse for
        # indexing.
        axvalues = [[i for i in range(s)] for s in shape] 
        cvalues = tuple(center)
        return_numpy = True
        
    elif isinstance(axes[0], tuple) and len(axes[0]) == 2 and isinstance(axes[0][0], str):
        # Named axes, `xarray` or `pandas` style. Each axis element
        # is either one `xarray` Axis object (with values in the .values
        # property) or a (name, values) axis tuple.
        # `center` is either a dict, or a list of (key, val) tuples.
        axvalues =  [ (x[1].values if hasattr(x[1], "values") else x[1])   for x in axes]
        cobj = dict(center)
        cvalues = tuple([cobj[x[0]] for x in axes])
        return_numpy = False
        
    else:
        # numpy array, explicit axes. Each axis is a 1D-iterable
        # and `center` is a 1D-tuple (one coordinate per axis).
        axvalues = axes
        cvalues = center
        return_numpy = True
    
    sqdist = np.abs(axvalues[0] - cvalues[0])**2
    for ax,c in zip(axvalues[1:], cvalues[1:]):
        sqdist = sqdist[...,None]
        sqdist = sqdist + np.ones(sqdist.shape) * (ax-c)**2
    
    if return_numpy:
        return sqdist
    
    return DataArray(data=sqdist, coords=[(x[0], x[1].values) for x in axes])
    
    
def stdwidth1d(data, center):
    '''
    Calculates the standard deviation a.k.a. "peak width" as
    the square-root of the
    [variance](https://en.wikipedia.org/wiki/Variance#Discrete_random_variable)
    of an experimental data array with respect to the value located in the
    same array at index position called `center`.

    For 1-dimensional arrays of discrete points, the variance is essentially
    defined as `(Xi-M)**2`, i.e. the squared sum of the differences of the
    poins `Xi` from a mean value `M`. It's a measure of how much each
    value differs from the mean.

    For N-dimensional data, the statistical reasoning is more complex and
    evolves around [the covariance matrix](https://en.wikipedia.org/wiki/Covariance_matrix).
    This function does *not* implement the covariance matrix version
    (serach for `stdvariance()` instead), but only the simple 1D version.
    If fed with a multiple N-dimensional input, it *still* only calculates
    one variance, which can essentially be interpreted as an "average variance"
    along all dimensions.
    
    Parameters:
    
      - `data`: an N-dimensional `numpy` array or an `xarray`.
        
      - `center`: if `data` is an numpy array, this is expected to be
        an N-tuple capable of indexing `data`. If `data` is an `xarray`,
        this must be a dictionary with dimensions as keys and positions
        as values, or a `(dim, value)` tuple.
      
    Returns: One single value, represending the sum of suqare distances
    weighted by the function value at the respective place, if
    `axis=None`. If `axis` is specified, then the result will be a
    1-dimensional array containing standard widths along `axis`.
    '''
    
    # Preparing the coordinate axis array of `data' -- for an xarray
    # we just build (name, values) tuples of its dims; for numy arrays,
    # we generate an integer index corresponding to the length.
    if hasattr(data, "coords"):
        axes = [(d, data.coords[d]) for d in data.dims]
    else:
        axes = [range(i) for i in data.shape]
        
    pos = center
        
    sqdist = sqdistance(*axes, center=pos)
    sqvar  = (sqdist * data / data.sum()).sum()
    
    return np.sqrt(sqvar)


def series_stdwidth1d(data, center, axis=0):
    '''
    Wrapper for `stdwidth1()` to act on a collection of data sets at once.
    #
    '''
    pass


def test_stdwidth1d_ndarray(test_ndarray):
    '''
    Test for stdwidth()
    '''

    center = (np.array(test_ndarray.shape)/2).astype(int)
    
    v = stdwidth1d(test_ndarray, center)

    # We're measuring the variance ("width") with regards to the
    # actual center of the data patch, so this should always essentially
    # be smaller than half the field.
    # This being random data, it *should* have a width larger than 1
    # grid point, though...
    assert (v <= np.abs(center.max()))
    assert (v >= 1.0)

    
def test_stdwidth1d_xarray(test_xarray):
    '''
    Test for stdwidth()
    '''

    # alias
    dta = test_xarray

    center = [ (x, dta.coords[x].values[0]+(dta.coords[x].values[-1]-dta.coords[x].values[0])/2 ) \
                     for x in test_xarray.dims ]
    
    v = stdwidth1d(test_xarray, center=dict(center))

    # We're measuring the variance ("width") with regards to the
    # actual center of the data patch, so this should always essentially
    # be smaller than the field.
    c_coord = np.array([dta.coords[x].values[-1]-dta.coords[x].values[0] for x in dta.dims])
    assert (v <= np.abs(c_coord.max()))


    
@pytest.fixture
def test_ndarray():
    ''' Returns a an ndarray of variable size and shape (i.e. dimensions) '''

    ## ...up to 7 dimensions should prove the point.
    ## Configurable for debugging.
    
    min_dims = 2
    max_dims = 2 #7

    min_size = 2
    max_size = 10 #20
    
    dims = min_dims + int(np.random.random() * (max_dims-min_dims))
    shp  = min_size + (np.random.random(dims) * (max_size-min_size)).astype(int)
    return np.array(np.random.random(shp.prod())).reshape(shp)


@pytest.fixture
def test_ndcenter(test_ndarray):
    ''' Returns a random 'center' coordinate that lies within the array '''
    return (np.random.random(len(test_ndarray.shape)) * test_ndarray.shape).astype(int)
    

def test_sqdist_ndarray(test_ndarray, test_ndcenter):
    '''
    Tests sqdistance() on an ndarray (i.e. no named axes)
    '''
    
    assert ((test_ndcenter < test_ndarray.shape).all())
    
    sqd1 = sqdistance(center=test_ndcenter, shape=test_ndarray.shape)
    
    axlist = tuple([range(s) for s in test_ndarray.shape])
    sqd2 = sqdistance(*axlist, center=test_ndcenter)

    # No distance can be longer than the diagonal across all dimensions
    assert (sqd1.max() <= (np.array(test_ndarray.shape)**2).sum())
    assert (sqd2.max() <= (np.array(test_ndarray.shape)**2).sum())
    
    # The square distance result has exactly one point for every input point
    assert (sqd1.shape == test_ndarray.shape)
    assert (sqd2.shape == test_ndarray.shape)
    
    assert ((sqd1[tuple(test_ndcenter)] == sqd2[tuple(test_ndcenter)]).all())
    
    # Distance at center is always zero
    assert((sqd1[tuple(test_ndcenter)] == 0).all())
    assert((sqd2[tuple(test_ndcenter)] == 0).all())
                                                                      
                                                                      
def test_sqdist_ndtuple(test_ndarray, test_ndcenter):
    '''
    Test whether sqtistance() accepts tuples as arguments
    '''
    axlist = tuple([range(s) for s in test_ndarray.shape])
    sqd = sqdistance(*axlist, center=tuple(test_ndcenter))
    assert (sqd.shape == test_ndarray.shape)
                                                                      
                                                                      
@pytest.fixture
def test_xarray(test_ndarray):
    ''' Creates an `xarray` with named axes based on a random ndarray '''
    
    axis_names = ['physics', 'trustee', 'warning', 'ethnic',
                  'tree', 'text', 'contraction', 'shell' ]
    
    return DataArray(data=test_ndarray,
                     coords=[np.array(range(s))*0.1-(s/20) for s in test_ndarray.shape],
                     dims=axis_names[:len(test_ndarray.shape)])
    

@pytest.fixture
def test_xcenter(test_xarray):
    '''
    Generates a random single-point center coordinate based on an
    array with named axes. The coordinate tuple has named dimensions.
    '''
    
    pos = {}
    for d in test_xarray.dims:
        ax = test_xarray.coords[d].values
        r = np.random.random()
        c = ax[0] + r*(ax[-1]-ax[0])
        pos[d] = c

    return pos


def test_sqdist_xarray(test_xarray, test_xcenter):
    '''
    Tests the sqdistance() function with xarray-like named axes.
    '''
        
    ## This is for more specific testing / debugging.
    #shape  = (5, 5)
    #data   = np.ones(shape)
    #axes   = [(np.array(range(s))-s/2)*0.1 for s in shape]
    #center = {'x': 0, 'y': 0 }
    #xdata  = DataArray (data=data, coords=axes, dims=['x', 'y'])
    
    center = test_xcenter
    xdata  = test_xarray
    
    sqd = sqdistance(*xdata.coords.items(), center=center)
    
    assert (sqd.max() <= (np.array(xdata.shape)**2).sum())
    
    # The square distance result has exactly one point for every input point
    assert (sqd.shape == xdata.shape)
    
    # Calculate distance (N-dim diagonal) between two ajacent axis coordinates.
    # This will give us a very crude estimate of an error measure which is
    # (1) smaller than a pixel, but
    # (2) still large enough such that an interpolation of the array value
    # at that specfic point will still fit well within that error value.
    
    pix_diag = np.sqrt(np.array([ (x[1].values[1]-x[1].values[0])**2 for x in xdata.coords.items() ]).sum())
    maxerr = pix_diag*0.5 / len(xdata.shape)
    
    #print ("### distances:\n", np.sqrt(sqd.values))
    #print ('### pixdiag:', pix_diag)
    #print ('### maxerr:', maxerr)
    #print ('### value(s) at center:', sqd.interp(center).values)
    #print ('### center coordinates:', center)
        
    # Distance at center is always zero
    assert( (sqd.interp(center) < maxerr).all())

    # Make sure sqdistance() accepts a list of tuples as center coordinates
    sqd2 = sqdistance(*xdata.coords.items(), center=center.items())
    assert ((sqd2 == sqd).all())

    # Make sure sqdistance() accepts a dict() center coordinates
    sqd3 = sqdistance(*xdata.coords.items(), center=dict(center.items()))
    assert ((sqd3 == sqd).all())

    # Make sure sqdistance() accepts a list of tuples as axes
    #sqd3 = sqdistance(*tuple( [(d,xdata.coords[d]) for d in xdata.dims] ),
    #                  center=center)
    #assert ((sqd2 == sqd3).all())




#
# Testing only below this point
#
import pytest
from io import StringIO

from nx5d.runlog import RunlogFile

def test_header_standard():
    '''
    Tests the "key: value" header setup. The "*_simple.log" files
    are straight-forward, no surprises; the regular files may contain
    empty lines and/or leading spaces and other non-breaking format
    "impurities". They all have to pass
    '''

    for fname in [
            './tests/test_data/runfile/run105_simple.log',
            './tests/test_data/runfile/run105.log' ]:
        r = RunlogFile(None)
        f = open(fname)
        result = r._consume_header(f)
        assert len(result[0]) == 17

        # These are just a few, feel free to extend:
        for i in [ "sample 4", "source", "sample slits" ]:
            assert (i in result[0].keys())

        assert int(result[0]["pulses per image"]) == 1000


def test_data_standard():
    '''
    Tests proper loading of data. We do this by initializing the
    object as-is, but we only care about integrity of the data
    part.
    '''
    for fname in [ ]:
        r = RunlogFile(fname)
        assert len(r.keys()) == 12


def test_header_broken():
    '''
    Loads test file(s) with broken headers / attrs and checks
    for proper error reporting.
    '''
    pass


def test_broken_data():
    '''
    '''
    pass



#!/usr/bin/python3

import fabio
import xrayutilities as xray
import numpy as np
import argparse
import experiment


def makeXray(experiment):
    '''
    Creates / returns an xrayutilities.Experiment() class according to
    dictionary experiment specification.
    '''
    # The experiment["detector"] object always has a tupe of 3 axes
    # to specify; this is because some xrayutilities modules, e.g.
    # the HXRD object, use specific definitions.
    # Those that are not defined in a particular setup are set
    # as 'None' at the axis level.
    # We need to filter those out before passing them on to QConversion.
    
    #detector_geometry = {
    #    "axes": [],
    #    "angles": []
    #}
    #
    #for ax,rot in ):
    #    if ax is not None:
    #        detector_geometry["axes"].append(ax)
    #        detector_geometry["angles"].append(rot)

    detector = [ x for x in filter(lambda x: x[0] is not None,
                                                zip(experiment["detector"]["axes"],
                                                    experiment["detector"]["angles"])) ]
    det_axes, det_angles = zip(*detector)
    
    print ("Detector axes:", det_axes)
    print ("Detector angles:", det_angles)

    qconv = xray.QConversion(sampleAxis=experiment["sample"]["axes"],
                             detectorAxis=det_axes,
                             r_i=experiment["sample"]["inplane-direction"],
                             en=experiment["xray-energy"])
    
    e = xray.io.EDFFile(filename)
    run15x = e.ReadData()

    exp = xray.HXRD(idir=experiment["sample"]["inplane-direction"],
                    ndir=experiment["sample"]["normal-direction"],
                    sampleor=experiment["sample"]["orientation"],
                    qconv=qconv)

    exp.Ang2Q.init_area(detectorDir1=experiment["detector"]["sensor"]["directions"][0],
                        detectorDir2=experiment["detector"]["sensor"]["directions"][1],
                        cch1=experiment["detector"]["sensor"]["center"][0],
                        cch2=experiment["detector"]["sensor"]["center"][1],
                        Nch1=experiment["detector"]["sensor"]["size"][0],
                        Nch2=experiment["detector"]["sensor"]["size"][1],
                        # either distance/chwidth...
                        distance=experiment["detector"]["sensor"]["distance"],
                        pwidth1 =experiment["detector"]["sensor"]["channel-size"][0],
                        pwidth2 =experiment["detector"]["sensor"]["channel-size"][1],
                        # or channels-per-degree
                        # ...,
                        detrot     =experiment["detector"]["angles"][2],
                        tiltazimuth=experiment["detector"]["angles"][1],
                        tilt       =experiment["detector"]["angles"][0]
                        )

    return exp
    


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--image", help="Process a single image file")
    parser.add_argument("-x", "--experiment", help="Use experiment definiton from YAML file")
    #parser.add_argument("nexus", help="Opens / processes data from a HDF5/Nexus file")
    args = parser.parse_args()    

    if args.image:
        print ("Loading data:", args.image)
        f = fabio.open(args.image)
        run15f = f.data        
    else:
        print("No data specified.")

    xray = None
    if args.experiment:
        print ("Loading experiment setup:", args.experiment)
        exp = makeXray(fromYaml(args.experiment))
        xray = makeXray(exp)
    else:
        print("No experiment file, using default KMC3 builtin")
        xray = makeXray(experiment.KMC3)

    
    qset = exp.Ang2Q.area(*(experiment["sample"]["angles"]),
                         *(detector_geometry["angles"]))
   
    print ("Q:", qset)

    for qi in qset:
        print ("Shape:", qi.shape)



#!/usr/bin/python3

import yaml

# Device / experiment geometry.
# The idea is to have this as a standard data format for any input
# data to the transformation algorithm.
#
# Ideally we'd separate this into strictly geometric data (i.e.
# "how the detector is built") and measurement data ("where are
# my angles at this particular moment"), but these data seem
# to be pretty mixed up in xrayutilities.
#
# As a consequence, we keep them together here, too.
# Perhaps a (future) system where several of these dictionaries
# can be cascaded can be put in place?
KMC3 = {

    # Energy of the X-ray beam in eV
    # XPP: ?
    # ESRF: ?
    "xray-energy": 9400.0,    

    
    "sample": {

        # Sample surface orientation in the same coordinate system as
        # the goniometer rotations are given (?)
        #  - det:
        #  - sam: ...
        #  - xyz+-...
        "orientation": 'x+', #"sam",

        # X-ray direction at zero angles (as seen from the sample coordinate system)
        "inplane-direction": (0, 1, 0),
        
        # Sample surface normal (_always_ perpendicular to inplane-dir -- why necessary?)
        # XPP: n/a
        "normal-direction": (0, 0, 1),

        # Definition of rotation angles (and direction) of
        # the sample with resepct to the lab coordinate system,
        # from outer to inner most circles.
        # XPP: n/a, 
        "axes": ('y+', 'z+', 'x+'),

        # Angles of sample positioning
        # XPP: (Sxx/positioners/Theta, Sxx/positioners/Chi, Sxx/positioners/Phi)
        # ESRF: (instrument/positioners/gonx|y|z?
        "angles": (0, 0, 0),
    },

    "detector" : {

        "sensor": {

            # Directions of the detector along width/height of the pixel plate
            # XPP: n/a
            "directions": ('x+', 'z+'),
        
            # Center channels (pixels) of the detector, i.e.
            # the point through which the X-ray goes at zero
            # angles, should probably be the same as the axis
            # of rotation (?). If there is no rotation axis,
            # then this is just the point on which the incoming
            # X-ray beam hits (right?)
            # XPP: n/a
            "center": (283, 178),

            # Number of channels (pixels) in each dimenion
            # XPP: n/a
            # ESRF: ?
            "size": (195, 487),

            # The pixel geometry is either specified as distance/pixel-size,
            # or as pixel-per-degree
        
            # Size of a detector channel (pixel) in the same
            # units as the distance (see below), and
            # distance of detector center from the sample in the
            # same units as the pixel size (see above)
            # XPP: n/a
            # ESRF: ?
            "channel-size": (1.5e-4, 1.5e-4),
            "distance": 1.0,

            # Numer of pixels (channels) per degree, in either direction
            # XPP: n/a
            # ESRF: n/a
            "chperdeg": None, # (43, 57)
        },

        # Definition of rotation angles (and direction)
        # of the detector, relative to the lab system, in
        # the order (Tilt, Azimuth, Rotation), where:
        #  - Tilt: The in-sample-plane rotation of the detector;
        #    None of the axis is an experimental constant/never
        #    used, and to be set to 0.
        #  - Azimuth: Rotation in the detector plane (2-Theta?)
        #  - Rotation of the sensor itself around its own normal
        #    axis (around the incident beam axis at zero angle).
        #    None if never used, set to 0.
        # XPP: n/a
        # ESRF: ?
        "axes": (None, 'z+', None),

        # Detector angles, tuple as specified by "rotations" above.
        # XPP: (None, 'Sxx/positioners/TwoTheta', None)
        # ESRF: (instrument/positioners/detx|y|z?
        "angles": (0, 0, 0),
    },

    "data": {

        # Which value to normalize data by
        # XPP: ?
        # ESRF: instrument/machine/current
        "normalization": 1.0,
    }
}

def fromYaml(path):
    '''
    Reads an experiment specification file (YAML) and returns
    an experiment dictionary, with values as above.
    '''

    x = yaml.load(path)



#!/usr/bin/python3

def stop_series(start=-200, shape=(8,10), step=2, pval=0):
    '''
    Creates a series of size NxM (for `shape=(N, M)`) values starting
    at `start` and with a step of `step`, but includes a "stop" value
    of `pval` every `period` entries.
    
    For example:
       stop_series(start=1, stop=9, period=3)
       
       -> `[1, 2, 3, 0,  4, 5, 6, 0,  7, 8, 9, 0 ]` 
       
    Returns: An array with lenth `(N+1) * M`, containing values with `step`
    increment starting from `start`, interseded with values of `pval`.
    '''
    ts  = start+np.array(range(0, (shape[1]*shape[0])))*step
    ts2 = ts.reshape(shape[::-1])
    cat =  np.concatenate((ts2, pval*np.ones([shape[1]])[:,None]), axis=1)
    
    return cat.flatten()


#!/usr/bin/python3

from silx.io import commonh5 as ch5
from numpy import array as np_array
from itertools import product as itr_product

from numpy import array as np_array

from xarray import DataArray, Dataset, concat

import logging

'''
Reading XFEL Data
=================

This module implements analysis and transformation routines which enable
use of the :class:`ScanReader` module with data acquired at the
[European X-ray Free Electron Laser (XFEL)](https://www.xfel.eu/).
It uses the [EXtra-data](https://pypi.org/project/EXtra-data/) library,
the official XFEL data extraction library.

Understanding the XFEL Data Format
----------------------------------

There is [good documentation available](https://extra-data.readthedocs.io/en/latest/)
for the EXtra-data package, also containing an introduction to
[the on-disk XFEL data](https://extra-data.readthedocs.io/en/latest/data_format.html).
A conceptual overview of how the data is organized can be inferred from
there.

Here we outline the main features of the XFEL data, and how they relate
to our data [concepts](concepts.md):

  - **run**: the main unit is a "run", which is a not further specified "series"
    of measurements. This is similar to our
    [scan](concepts.md#scan_streak_frame), 
    but with the notable difference that a *scan* consists of a series
    of simlar *streaks*, while an XFEL "run" bears no such restrictions.

  - The datapoints themselves is organized in "detector data" and
    "other data", similarly to the *measurement* and *positioners*
    we've come to know from the ESRF falvors of Nexus.

  - A "run" may be broken down into "sequences". There is apparently
    no semantic restriction on what goes into a sequence, so it can't
    be related to any of our *frame* or *streak* concepts. Also,
    different sequences from different acquisition devices don't
    necessarily align, which is a requirement for our frames and
    streaks.

  - **train**: this is the usefully organized sub-unit of data acquisition;
    it holds all the data points (a.k.a. *frames*, in nx5d lingo) that
    come with a pulse train. (A pulse train is one "shot" of the XFEL, taking
    place within a a significant fraction of a second, and containing
    up to several hundreds of pulses at once.)

    Each of the instruments saves their data according to the current
    *train ID*, which is unique. Not every instrument does, however, necessarily
    save any data for each train, and those that do, don't necessarily save
    only a single data point; in fact, many save several, e.g. one data point
    for each pulse. The number of data points one specific instrument
    saves may also vary from one train to the next.

    As such, trains do not neatly align with our initial idea of *frames*,
    but don't outright contradict it.


Data Adapter Strategy
---------------------

Analyzing XFEL data is a two-step process:

  1. Data needs to be extracted from the pecular XFEL storage
     and translated into a more "handy" stucture, imitating
     the ESRF Nexus flavor nx5d is expecting. This can be
     done abusing the `ScanReader` infrastructure, but needs
     specific HDF5-like interfaces around EXtra-data's API elements,
     primarily the
     [`DataCollection`](https://extra-data.readthedocs.io/en/latest/reading_files.html)
     family of objects and generators.

  2. Standard nx5d tools can be applied to the output of step 1
     to further process and ingest
     data according to standard nx5d methods for specific experiments,
     e.g. the [X-ray diffraction]() module. This is also done using
     `ScanReader`, but in its own element. For this, of course,
     the output of step 1 needs to be a HDF5-like object itself.

So in essence, producing scientific insight from XFEL data involves
a stacked use of `ScanReader`. This module implements the necessary
HDF5-like input classes and output composers to magic the raw XFEL
produced data into something that "off-the-shelf" experimental analysis
modules of nx5d can make sense of.

Data Set Addressing
-------------------

Similarly to the way nx5d expects it, the XFEL data storage system
offers device labels to address the data. The main difference is that
not all devices go in-sync when *collecting* data: some collect
one datapoint per pulse train (but on a different pulse than another),
others collect several at different times, and yet others don't collect
any data at all.

In the first phase -- extracting and preparing -- we make it the
user's (or: higher API level's) responsibility to address only those
devices that have a similar storage scheme, e.g. the 16 modules
that make up one particular detector. To simplify and encourage
"correct behavior", addressing scheme will have syntactic sugar
to address a whole class of devices at once (e.g. selecting
`detector="AGIPD1M-1"` instead of `("MID_DET_AGIPD1M-1/DET/*CH0:xtdf")`),
in addition to regular, completely free key addressing.

This in turn also means that we are offering a list of special XFEL
devices, detectors in particular, with their naming scheme and data
processing particualrities. Beyond addressing, this is important for
a variety of reasons: XFEL detectors are not single pieces of equipment,
they are (sometimes?) made up of distinct components -- e.g. up to 16
individual chips. These have a non-trivial geometry and spatial relation
to one another, and need supplementary data (and API elements from
EXtra-data) in order to assemle a full, "regular" x-ray diffraction
pattern picture.
'''

class ExdatLazyDataset(ch5.LazyLoadableDataset):
    ''' Node to lazily load `extra_data.KeyData` datasets.
    '''
    def __init__(self, keydata, name, parent=None, attrs=None, array_like='dask_array'):
        super().__init__(name, parent, attrs)
        self._kdata = keydata
        self._array_like = array_like

    def _create_data(self):

        # FIXME: Loading a dask_array() from _kdata and using it by default
        # is a pointless exercise in burning CPU cycles.
        # This is because we're most likely going to slice the dataset anyway
        # (using ScanReader.streaks()), and we're  typically going to load
        # a lot less than the whole data into memory.
        # Which means that we're likely to not have to _ever_ use dask_array(),
        # unless for the seldom cases where we do arithmetics on the dataset
        # itself.
        #
        # Typical usage scenario will rather be something like "dataset[xxx]",
        # and xxx will be small enough to be returned as a ndarray on virtually
        # every occasion.
        #
        # This needs to be fixed.
        #
        # Not sure how to do that, though... most likely we need to extend
        # __getitem__() below to actually intercept slicing requests on
        # dask arrays, and peform _kdata.ndarray(roi=(...)) selections when
        # that happens. This will allow us to have fast ndarray-based data
        # management when loading small(-ish) streaks.
        #
        # On the other hand, the danger is that the user is stupid and will
        # store all the streaks before starting to process them... which
        # will fill up the memory, again.
    
        
        d = getattr(self._kdata, self._array_like)()
        logging.debug("Creating data: %s#%s with shape %r" % \
                      (self._kdata.source, self._kdata.key, d.shape))
        return d

    def __getitem__(self, item):
        ''' Overriding __getitem__() from silx's LazyLoadableDataset.

        The guys over at ESRF were a little bit too ambitious, respectively lacking in
        phantasy :) and they're checking whether the dataset is an ndarray instance.
        Which, of course, it isn't, since we're (sometimes?) using dask arrays.
        So we need to catch for this here. So what we're doing is catching the
        case where data is a container (with "__len__"), but not an ellipsis or
        a tuple, and instead of raising an error, we're passing the call on
        to data's __getitem__().

        This will probably break at some point, and then some poor soul will have
        to cross-check again with the silx's commonh5.LazyLoadableDataset sources
        and see how to work around that. Good luck!
        '''
        if hasattr(self._get_data(), "__getitem__"):
            return self._get_data().__getitem__(item)
        else:
            return super().__getitem__(item)


    @property
    def shape(self):
        ''' Same problem as __getitem__(): overzealous superclass discards data if it's not numpy.
        '''
        return self._get_data().shape \
            if hasattr(self._get_data(), "shape") \
               else super().shape

    @property
    def size(self):
        ''' Same problem as __getitem__(): overzealous superclass discards data if it's not numpy.
        '''
        return self._get_data().size \
            if hasattr(self._get_data(), "size") \
               else super().size


    def __len__(self):
        ''' Same problem as __getitem__(): overzealous superclass discards data if it's not numpy.
        '''
        return self._get_data().__len__() \
            if hasattr(self._get_data(), "__len__") \
               else super().__len__()
    
        

class ExdatSourcesGroup(ch5.Group):
    '''Mock HDF5 node for a subset of `extra_data.DataCollection` keys.
    '''
    def __init__(self, name, parent, selection, array_like='auto'):
        super().__init__(name=name, parent=parent)
        self.sel = selection

        for src in self.sel.all_sources:
            snode = self._ensure_group(self, src)
            logging.debug("Have node: %s" % snode.name)
            
            for key in self.sel.keys_for_source(src):
                kdata = self.sel[src,key]

                if array_like == 'auto':
                    numpts = np_array(kdata.shape).prod()
                    array_like = 'dask_array' if numpts > 1e6 else 'ndarray'
                
                snode.add_node(ExdatLazyDataset(kdata, name=key, parent=snode,
                                                array_like=array_like))
                
                logging.debug("  Dataset: %s" % key)
                


    def _ensure_group(self, parent, source_path):
        '''Makes sure all path components of source_path exist.

        Returns:
            The corresponding group node object.
        '''
        parts = source_path.split('/') if isinstance(source_path, str) else source_path
        me = parts[:1][0]
        them = parts[1:] if len(parts)>1 else None
        
        if me not in parent:
            parent.add_node(ch5.Group(name=me, parent=parent))

        return parent[me] if them is None \
            else self._ensure_group(parent[me], them)
        
        
class ExdatH5(ch5.File):
    '''Generic class to offer HDF5-like access to a `DataCollection` object.

    This implementation is not necessarily fully h5py-API compatible. It contains
    just enough of a mock up to be able to work nicely with nx5d.scan.ScanReader.
    
    Currently the HDF5 mock is flat, i.e. no groups, just the base data sets.
    '''

    def __init__(self, dsel, array_like='auto'):
        '''Initializes the ExdatH5 instance.
        
        Args:
            dsel (extra_data.DataCollection): the data collection (selection) to expose.
        
            array_like: One of "numpy", "dask" or "auto", representing the default
              array mechanism to use when retrieving data from DataCollection.
              "numpy" is strongly suggested if the data will fit in memory,
              since it's considerably faster. However, "dask" will provide you
              with distribuited arrays that work across a large number of nodes,
              should you need it. The default "auto" will always try "numpy"
              first, and fall back to "dask" if that fails.
        '''
        super().__init__()

        self.add_node(ExdatSourcesGroup(name="measurement", parent=self,
                                        selection=dsel.select([
                                            (k, '*') for k in dsel.all_sources
                                        ]),
                                        array_like=array_like))
        
        #self.add_node(ch5.Group(name="instrument", parent=self))
        #self['instrument'].add_node(ExdatGroup(name="positioners", parent=self['instrument'],
        #                                       selection=None))
        
        self.dsel = dsel

        # could split these in instrument_sources and control_sources.
        #self.datasets = {k:self.dsel.get_virtual_dataset(k) \
        #                 for k in self.datc.all_sources}

    def __enter__(self, *args, **kwargs):
        ''' Overriding the `silx.io.commonh5.File` enter/exit mechanism.
        
        ScanReader depends on using nested with-guards (enter/exit), i.e.
        entering/exitting multiple times on the same object.
        This is an extension to do just that.
        '''
        if not hasattr(self, "_enter_cnt") or self._enter_cnt == 0:
            self._enter_data = super().__enter__(*args, **kwargs)
            self._enter_cnt = 0
        self._enter_cnt += 1
        return self._enter_data


    def __exit__(self, *args, **kwargs):
        if self._enter_cnt > 0:
            self._enter_cnt -= 1
        if self._enter_cnt == 0:
            super().__exit__(*args, **kwargs)


from extra_data.components import AGIPD1M
            
''' Quick'n Dirty database of XFEL detectors and their data.

This is what we need to convert detector tiles into XRD images.
'''
detectors = {
    'AGIPD1M-1': {
        'selection': [
            ('MID_DET_AGIPD1M-1/DET/*CH0:xtdf', '*'),
            ('MID_AGIPD_MOTION/MDL/DOWNSAMPLER', '*ActualPosition.value'),
            ('MID_EXP_EPIX-1/DET/RECEIVER:daqOutput','data.image.pixels'),
            ('SA2_XTD1_XGM/XGM/DOOCS:output','data.intensityTD'),
            ('MID_EXP_SAM/MDL/DOWNSAMPLER','ssryActualPosition.value'),
        ],

        'data_template': {
            'tmp':  tuple([ '@{measurement}/MID_DET_AGIPD1M-1/DET/%dCH0:xtdf/'
                              'image.data' % i \
                              for i in range(1) ]),

            'tiles': lambda h5, nodeOnly, slicer: \
            AGIPD1M(h5.dsel).get_dask_array('image.data').transpose('train_pulse', 'module', 'dim_0', 'dim_1')[slicer],
            
            'motors': tuple([ ('@{measurement}/MID_AGIPD_MOTION/MDL/DOWNSAMPLER/q%dm1ActualPosition.value' % i,
                               '@{measurement}/MID_AGIPD_MOTION/MDL/DOWNSAMPLER/q%dm2ActualPosition.value' % i)
                              for i in range(1,5) ]),

            'angles': {
                'theta': '@{measurement}/MID_EXP_SAM/MDL/DOWNSAMPLER/ssryActualPosition.value',
            },

            'norm': {
                'epix': '@{measurement}/MID_EXP_EPIX-1/DET/RECEIVER:daqOutput/data.image.pixels',
                'xgm':  '@{measurement}/SA2_XTD1_XGM/XGM/DOOCS:output/data.intensityTD',
            }
        },
    },
}


class DetectorTiler:
    ''' Combines images of individual XFEL detector modules into an 'XRD-like detector image.'

    This is a higly specialized class which generates exacly *one* kind of
    image, in exactly *one* kind of way.  [FIXME: document how!]
    '''

    def __init__(self, geom_class=None, geom=None,  mask=None, tiles=None,
                 norm=None, motors=None, passthrough=True, **passthrough_data):
        ''' Initializes the tiler with a dataset and all metadata necessary for tiling.
        '''

        # This is the trick part that we'll need to fix: it's a major pain in the
        # ass to tile together the detector images from its individual modules.
        # The best thing is to use the desginated detector object (e.g. AGPID1M(...)),
        # but this means that this needs to be done on a HDF5-like level.
        self.tiles  = tiles

        
        self.norm   = norm
        self.motors = motors
        self.geom   = geom
        self.mask   = mask
        self.geom_class = geom_class        
        self.passthrough_data = passthrough_data.copy() if passthrough else {}

        try:
            epix = self.norm['epix'].compute().mean(axis=(1,2))
            xgm = self.norm['xgm'].compute()[:,0]
            print ("EPIX:", epix.shape)
            print ("XGM:", xgm.shape)
            self.valid_data_selector = \
                self._norm_filter(epix) * \
                self._norm_filter(xgm)
        except Exception as e:
            logging.error("While calculating norm: Oops, %r" % str(e))


    def _geom_from_motors(self, q1m, q2m, q3m, q4m):
        ''' Returns a geometry object from the specified quadrant motor positions.
        '''
        quad_pos = (
            (-542 + 0*q1m[0],  660 + -5*q1m[1]),
            (-608 + 0*q2m[0],  -35 +  5*q2m[1]),
            ( 534 + 0*q3m[0], -221 +  5*q3m[1]),
            ( 588 + 0*q4m[0],  474 + -5*q4m[1]),
        )
        
        #print ("Quad positions:", [(i[0].compute()[()], i[1].compute()[()]) for i in  quad_pos])
        
        return self.geom_class.from_quad_positions(quad_pos=quad_pos)
        

    def _norm_filter(self, data):
        ''' Filter normalization values.

        Filtering criteria are:
        
           - exclude everything that stands out by more than one order
             of magnitude from the ful dataset
        
           - then exclude everything that stands out by more than half
             from the remaining dataset

        Returns:
            A data selector (boolean array slicer) with valid data,
            based on the fitlering criteria.
        '''
        norm_data = data.mean()/data
        sel_data = (norm_data>0.1) * (norm_data<10)
        norm_data = data[sel_data].mean()/data
        sel_data = (norm_data>0.5) * (norm_data<2)
        return sel_data
        
    
    def __call__(self, pulse_slicer, mask=None, tiles=None, **pulse_reducers):
        '''Returns the sum of full detector images specified by pulse_list

        Args:
        
            pulse_slicer: Slicer of pulses to select (e.g. `numpy.s_[2,3,17]`)
        
            mask: Array of shape `(N, width, height)`, where `N` is the number
              of detector modules (typically 16), and `width` and `height` are
              the geometry of the detector module images, as they come in
              the tiles argument of `__init__()`.
        
            tiles: Callable for reducing multiple detector tile(s)
              (i.e. as many as there are pulses) into one. The callable may
              receive either a 3D (shape: `(N, width, height)`) or a 4D
              xarray (shape: `(P, N, width, height)`), where `P` is the number
              of pulses, `N` the number of detector tiles, and `width` / `height`
              are the image sizes of the tiles themselves. This depends on how
              `pulse_slicer` is set.
              It is expected to return a 3D array `(N, width, height)` by custom
              transformation (e.g. summing up data along the first dimension).

            pulse_reducers: Hashable which contains a data key (as passed
              to `__init__()`) as key, and a pulse-reducing callable as value.
              The pulse-reducing callable has a similar use as `tiles`, i.e.
              reducing the data to one single instance. The callable is passed
              the entire dataset with the corresponding name and is expected
              to return the reduced data payload (either as array or as a scalar).
              Every data set for which a pulse-reducer is present, is also
              returned in the output dictionary.

        Returns:
            A dictioanry containing at least the following keys:
        
            - "image": this is the composed detector image made up
              of all the tiles.

            - "metrics": a Python `dict` with as many of the fields that need to
              to be passed on for XRD Q-space transformation (e.g. inside the
              "setup" parameter of the `nx5d.xrd.LazyQMap` object) as can be
              computed from the tile data.
              Note that some of the metrics values might be wrong, or require
              manual offsetting to account for alignment issues.
        
              Typically, these fields are:
        
              - "imageChannelSize": the virtual "pixel size" of the full detector,
                inferred from the x, y, z return values
                of `.get_pixel_positions()` of a suitable geometry class.

              - "imageSize": the size of the full 'virtual' detector image generate,
                in pixels

              - "imageCenter": the center pixel of the 'virtual' detector image,
                as received from `.position_modules()` of a suitable geometry clas.

              - "imageDistance": the detector distance from sample, inferred from
                the z coordinates. (ACHTUNG, this is particularly likely to be wrong
                in many occasions, e.g. if the coresponding motors haven't been
                initialized properly.)

            Additional keys are generated for every dataset for which there is an
            entry available in `pulse_reducers`, with the corresponding output.
            For instance: requesting to generate a $\theta$ angle by picking the
            first one from a list of such angles would imply an invocation like
            this: `.(s_[...], angles=lambda a: {k:v[0] for k,v in a.items()})`.
        '''

        # FIXME: this lambda assumes a structure different than the one
        # in the "motors" keys (this assumes a single x/y set for one quadrant,
        # while the dataset has all 4 quadrants -- this needs to be changed)
        motors = pulse_reducers.get('motors', lambda m: m[0])
        
        if tiles is None:
            tiles = lambda t: t.sum("train_pulse") if len(t.shape)>3 else t

        if mask is None:
            mask = self.mask if self.mask is not None else 1
             
        motorset = [(motors(x), motors(y)) for x,y in self.motors]
        geom = self._geom_from_motors(*tuple(motorset))

        tileset = tiles(self.tiles[pulse_slicer])
       
        xrd_image, xrd_center = geom.position_modules(tileset * mask)
        x, y, z = geom.get_pixel_positions().T

        result = {
            "image": DataArray(data=xrd_image,
                               coords={"x": np_array(range(xrd_image.shape[-2])) - xrd_center[-2],
                                       "y": np_array(range(xrd_image.shape[-1])) - xrd_center[-1]}),
            "metrics": {
                "imageChannelSize": ((x[-1,0,0] - x[0,0,0]) / x.shape[0],
                                     (y[0,-1,0] - y[0,0,0]) / y.shape[1]),
                "imageSize": xrd_image.shape,
                "imageCenter": tuple(xrd_center),
                "imageDistance": z.mean()
            }
        }

        result.update({k:pulse_reducers[k](self.passthrough_data[k]) \
                       for k in pulse_reducers.keys()})

        return result



#!/usr/bin/python3

'''
This module contains code to export various data formats
parsed by EDFnx (EDF format, Runlog format, ...) to HDF5/Nexus
files.
'''

from nx5d import runlog

import h5py
import numpy
import re
import fabio
from os import path

def _np2nx_type(np_type):
    '''
    Returns a HDF5 data type string, given a numpy type.
    '''
    if 'U' == np_type.kind:
        return numpy.dtype(np_type.str.replace('U', 'S'))
    else:
        return np_type
    

def _c2n_require_group(h5root, h5path, autodefault=True):
    '''
    copy2nexus helper - calculates a HRD5 path and creates
    the corresponding Nexus group(s). Returns a h5py HDF group object.
    Parameters:
      - `h5root`: Top level object, owner of the group, relative starting
        point for `h5path`.
      - `h5path`: String path description, possibly with Nexus
        group types.
      - `autodefault`: If set to `True`, each level receives a
        "@default" HDF5 attribute for the subsequent path element
        (Nexus standard).
    '''
    pathchain = []
    for i in h5path.split('/'):
        pathchain.append(list(i.split(':')))

    for pair in pathchain:
        if len(pair) == 1 and len(pair[0]) > 0:
            pair.append("NXentry" if pair != pathchain[-1] else "NXdata")
    
    base = h5root
    sub = None

    # The 1st element in the pathchain will be an empty string if
    # the path starts with '/'. Need to jump that.
    for d in (pathchain if len(pathchain[0][0]) > 0 else pathchain[1:]):
        sub = base.require_group(d[0])
        if 'NX_class' not in sub.attrs.keys():
            sub.attrs['NX_class'] = d[1]
        if 'default' not in base.attrs.keys() and autodefault:
            base.attrs['default'] = d[0]
        base = sub

    # full path: "/".join([i[0] for i in pathchain])
    # pathchain: pathchain

    return sub or base


def copy2nexus(dataobj, h5obj=None, h5group=None,
               h5path='/entry:NXentry/data:NXdata',
               nxsignal='auto',
               nxaxes='auto',
               nxautodefault=True,
               typeConverter=_np2nx_type):
    '''
    Takes an EDFnx data object (`dataobj`) and inserts it into a
    Nexus-compliant HDF5 file.

    The HDF5 file, and position (path) therein of the new data, can
    be specified in two ways:
      - either by specifying a HDF5 file object (via `h5obj`),
        and optionally an (absolute) path therein,
      - or by specifying a HDF5 group (via `h5group`), and
        optionally a (relative) path therein.

    The parameters are as follows:
      - `dataobj`: The EDFnx data object.
      - `h5obj`: The HDF5 object to contain the data. Default is `None`,
        but either this or `h5group` must be specified. If both are
        specified, this consists an ambiguity and an exception is raised.
      - `h5group`: The HDF5 object to contain the data. Default is `None`,
        see `h5obj` for details.
      - `h5path`: Path within the HDF5 file to contain the data. The format
        of the path is `[/]group[:type][/subgroup[:subtype]][/...]`. `type`
        is typically one of the
        [Nexus group classes](https://manual.nexusformat.org/classes/index.html#all-class-definitions).
        The leading `/` is optional when `h5obj` is set. If the class is not
        specified for a particular group path component, all groups are
        assumed to be `NXentry`, except for the last one, which is `NXdata`.
        If the group already exists, it's not being created.
      - `nxsignal`: The default data signal column; this is set as the default
        of the corresponding NXdata group. Special options include `None`
        (i.e. don't set the default), or 'auto' (the default), meaning
        that the algorithm will choose one by some meaningless heuristic.
      - `nxaxes`: Similar to `nxsignal`, but will set the axis instead of the
        signal attribute.
      - `nxautodefault`: Auto-sets the default descendency path (Nexus lango:
        `"@default"` attribute) for every path element.
      - `typeConverter`: This is a function that takes a `numpy.dtype` as an
        argument and returns another `numpy.dtype`. It is used to transform
        data types (mostly strings) since the HDF5 format chokes on specific
        kinds of data (e.g. unicode strings).
    '''

    grp = _c2n_require_group(h5group if h5group is not None else h5obj,
                             h5path, autodefault=nxautodefault)

    for k in dataobj.attrs.keys():
        grp.attrs[k] = dataobj.attrs[k]


    # Handle explicit nxsignal/nxaxes.
    # We don't check whether the columns really exist.
    if nxsignal not in [None, 'auto']:
        grp.attrs['signal'] = nxsignal
    if nxaxes not in [None, 'auto']:
        grp.attrs['axes'] = nxaxes
        
    for dname in dataobj.keys():
        
        a = dataobj[dname]
        dtype = typeConverter(a.dtype)
        
        grp.require_dataset(name=dname, data=a.astype(dtype), shape=a.shape, dtype=dtype)

        # Handle nxaxes/nxsignal as the 1st, respectively 2nd unused
        # floating-point data type.
        if a.dtype == numpy.dtype(float):
            if 'axes' not in grp.attrs.keys() and nxaxes == 'auto' and nxsignal != dname:
                grp.attrs['axes'] = dname
            elif 'signal' not in grp.attrs.keys() and nxsignal == 'auto' and nxaxes != dname:
                grp.attrs['signal'] = dname

                

def test_c2n_reqgroup():

    # The Nexus standard expects to have at least one NXentry-class folder,
    # with at least one NXdata-class subfolder. Obviously, some of these
    # paths make this impossible. 
    for p in [ "/entry/through/the/gates/of/hell",
               "/entry:NXentry/through:NXentry/the/gates/of/hell:NXdata",
               "entry/through/the/gates/of/hell",
               "",
               "foo/bar",
               "foo" ]:
        
        h5 = h5py.File('jinkies', driver='core', backing_store=False, mode='w')
        grp = _c2n_require_group(h5, p)
        
        assert grp.parent is not None
        
        # For easier testing, longer paths all have the same form (see above)
        if len(grp.name) > 10:
            assert grp.name == "/entry/through/the/gates/of/hell"

        # For paths with insufficient subgroups, the NX_class attribute
        # is botched up. Make sure we don't test that in those cases.
        if grp.name != '/':
            assert grp.attrs["NX_class"] == "NXdata"
        if grp.parent.name != '/':            
            assert grp.parent.attrs["NX_class"] == "NXentry"
            
        h5.close()
    
    
def test_c2n_runlog():
    '''
    Tests the generation of a HDF5/Nexus file.
    This doesn't actually validate the Nexus file itself (as this
    depends a lot on the nx* parameters within the `make_nexus()`
    function call). It just makes sure the code runs and essentially
    contains the data as expected
    '''

    r = runlog.RunlogFile("./test_data/run105.log")
    h5 = h5py.File('jinkies', driver='core', backing_store=False, mode='w')
    
    copy2nexus(r, h5obj=h5)

    h5data = h5["/entry/data"]
        
    for k in r.keys():
        assert h5data[k].shape == r[k].shape

    for k in r.attrs.keys():
        assert r.attrs[k] == h5data.attrs[k]


def zip2nexus(logobj, dirbase,
              h5obj=None,
              h5pathfmt_data='/scan-{__scan__:04}:NXentry/data:NXdata',
              h5pathfmt_vars='/scan-{__scan__:04}:NXentry/vars:NXinstrument',
              scanFormatter=lambda d: int(d["__index__"])+1,
              datapathFormatter=lambda d: u'%s.edf' % d["file"],
              datanameFormatter=lambda d: u'edfmap',
              typeConverter=_np2nx_type):
    '''
    Takes a RunlogFile object (or any EDFnx compatible) and integrates all
    the corresponding external data, and the Logfile data lines, into
    the specified HDF5/Nexus target (via `h5obj` and `h5pathfmt`).

    The ideas behind this are the following assumption:
    
      - The "real" experimental data is not contained within
        the Logfile (i.e. `logobj`) but in external files
        (e.g. EDF or TIFF files)
    
      - A "logfile" is actually only an experimental summary, holding
        named instrumental parameters; the data column (names) within a
        Logfile are just keys for named parameters, and each line
        within the data vectors represent setup values for one specific
        scan of the experiment.

    Following this, `zip2nexus` goes through each data line and loads
    the corresponding external data, saving it into the HDF5 object
    at its specific place. The metadata as specified by the columns
    of the data line are stored, together with the external data,
    as attributes of that specific NDF5 subgroup.

    Parameters:
    
      - `logobj`: Logfile data object, e.g. a `RunlogFile()` instance

      - `dirbase`: The base path (i.e. path without the file name component)
        of the file on which `logobj` is based. This is used as the base
        path for the `datapathFormatter`.

      - `h5obj`: HDF5 object that will be the root of the newly injected
        data. It may contain other data. It is expected to implement
        a `h5py` compatible interface to handle datasets and subgroups
        (i.e. `require_dataset()`, respectively `require_group()`)

      - `h5pathfmt_scan`: Path name template for the scan top folder,
        relative to the root of `h5obj`.
        The scan number will be inserted at the position of the `{__scan__}`
        parameter. All the header keys of the `RunlogFile` keys are also
        available as parameters, aswell as `{__index__}`, the iteration
        index of the data lines starting with 0.
        Default is `/scan-{__scan__}:NXentry`.

      - `h5pathfmt_data`: Template for the data folder, relative to the
        `h5obj` root. Same variables as for `h5pathfmt_scan`
        are available, default is `/scan-{__scan__}:NXentry/data:NXdata`.

      - `h5pathfmt_vars`: Template for the instrumental setup variables
        folder, relative to the root of `h5obj`.
        Same variables as for `h5pathfmt_scan`
        are available, default is `/scan-{__scan__}:NXentry/vars:NXinstrument`.

      - `scanFormatter`: This is a function receiving a dictionary as input
        (all columns of the `logobj` data, with the values of the
        corresponding data line as data, and `__index__` as the line index)
        and expected to output a numerical scan key (e.g. int(5) for the
        5th scan).

      - `datapathFormatter`: This is a function receiving a dictionary as input
        (all columns of the `logobj` data, with the values of the
        corresponding data line as data, and `__index__` as the line index,
        and `__scan__` as the scan key) and expected to output a file path
        relative to the folder of the logfile, as string.

      - `datanameFormatter`: This is a function receiving a dictionary as input
        (all columns of the `logobj` data, with the values of the
        corresponding data line as data, and `__index__` as the line index,
        and `__scan__` as the scan key) and expected to output a name for
        the dataset.

      - `typeConverter`: This is a function that takes a `numpy.dtype` as an
        argument and returns another `numpy.dtype`. It is used to transform
        data types (mostly strings) since the HDF5 format chokes on specific
        kinds of data (e.g. unicode strings).
    '''

    header = logobj.keys()
    
    for index, line in enumerate(zip(*(logobj.values()))):
        values = {k: v for (k,v) in zip(header, line) }
        values['__index__'] = index
        values['__scan__'] = scanFormatter(values)

        # Saving the data.
        datapath = datapathFormatter(values)
        name     = datanameFormatter(values)
        
        fab = fabio.open(path.join(dirbase, datapath))
        
        h5datagrp = _c2n_require_group(h5obj, h5pathfmt_data.format(**values))
        h5data = h5datagrp.require_dataset(name=name,
                                           data=fab.data,
                                           shape=fab.data.shape,
                                           dtype=fab.data.dtype)
        h5datagrp.attrs["signal"] = str(name)
        h5datagrp.attrs["axes"] = [ ".", "." ]
        
        h5varsgrp = _c2n_require_group(h5obj, h5pathfmt_vars.format(**values),
                                       autodefault=False)

        # Copying EDF file metadata into the HDF5 data folder. For
        # conveninence, trying to guess some of the field types
        # (int and float). Everything else stays 
        for (k,v) in fab.header.items():

            final_v = v
            try:
                final_v = int(v)
            except ValueError:
                try:
                    final_v = float(v)
                except ValueError:
                    final_v = v

            old_type = type(final_v)
            old_dtype = numpy.dtype(old_type)
            new_dtype = typeConverter(old_dtype)
            new_type = new_dtype.type                    
            data_array = numpy.array([final_v], dtype=new_dtype)

            try:
                existent = h5datagrp[k]
                if existent.shape != data_array.shape:
                    raise RuntimeError("Request to update key %s be updated with "
                                       "incompatible data %r" % (k, data_array))
                    
                existent[:] = data_array[:]
            except KeyError:
                h5datagrp.create_dataset(name=k, data=data_array)
                

        # Saving the runlog metadata
        for k,v in values.items():
            if k in [ '__index__', '__scan__' ]:
                continue

            # Nexus/HDF5 has problems with u'' strings. Recycling
            # the typeConverter(), which works from/to numpy.dtype().
            # Yes, this could be a oneliner, but breaking it down makes
            # it easier to understand.
            old_type = type(v)
            old_dtype = numpy.dtype(old_type)
            new_dtype = typeConverter(old_dtype)
            new_type = new_dtype.type
            data_array = numpy.array([v], dtype=new_dtype)

            try:
                existent = h5varsgrp[k]
            except KeyError:
                h5varsgrp.create_dataset(name=k, data=data_array)
        
    

def test_z2n_runlog():
    '''
    Tests the `zip2nexus()` function on `RunlogFile` data.
    '''

    rlog = runlog.RunlogFile("./test_data/short_run04.log")
    h5 = h5py.File('jinkies', driver='core', backing_store=False, mode='w')

    # Need to change the datapathFormatter here because the logfile+edf
    # combination in this particular example is actually a scam (we
    # removed all the actual data from EDF files and replaced it with
    # a 10x10 numpy array of undefined data).
    zip2nexus(rlog, "./test_data", h5,
              datapathFormatter=lambda d: "short_%s.edf" % d["file"])


    no_scans = 0
    for i in rlog.values():
        no_scans = len(i)

    # Just a quick n dirty test if the number of data files pans out.
    assert no_scans == len(h5.keys())

    # (If we've made it this far, then at least the test didn't raise
    # any exceptions :-)
    #
    # Feel free to add more assertions and/or tests as you find more
    # corner cases to be handled.
    

if __name__ == "__main__":

    from sys import argv
    from os import path
    from sys import exit

    if len(argv) < 3:
        print ("Usage: %s <logfile> <hdf5file>" % argv[0])
        exit (-1)

    rlog = runlog.RunlogFile(argv[1])
    h5 = h5py.File(argv[2], mode='a')

    zip2nexus(rlog, path.dirname(argv[1]), h5)

    h5.close()



#!/usr/bin/python3

#
#        nx5d - The NX5 Duct Tape
#        Copyright (C) 2022-2023 Florin Boariu
#
#        This program is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        This program is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

#from xrayutilities.io.spec import SPECFile

from silx.io.spech5 import SpecH5, SpecH5LazyNodeDataset, SpecH5NodeDataset
from silx.io.commonh5 import Group as H5Group, SoftLink as H5SoftLink
from h5py import File as H5File
from os import path
import tifffile
import numpy as np

'''
Provides a series of objects to mock an "ESRF dialect" of a HDF5-based Nexus
data file for X-ray diffraction experiments that are usable with `source.DataSource.

The "ESRF dialect" is essentially this:

  - There is 1D or 2D detector data ("images" or "frames").

  - The data is organized in "scans", i.e. images are oranized in series.
    One scan typically consists of several images with the same hard characteristics
    (i.e. same geometry), but taken under systematically differing external parameters:
    e.g. by "scanning" one or several parameter devices (angle? time delay?) and
    obtaining an image at each of the scan points.

  - The data format where the images are stored is a 3-dimensional block of NxWxH,
    where 'WxH' is the width/height of one frame, and 'N' is the number of frames
    in a scan. For a scan performed along a single paramter, each subsequent frame
    corresponds to a different parameter instance; for a scan taken along 2 parameters
    (e.g. inside a 2D parameter space), the scans are interleaved: there are actually
    (n*m)xWxH frames. This is still a 3D dataset with images of geometry WxH, but
    the first "chunk" of 'n' images are scanned along the first paramter and a specific
    'm' value (i.e. all have the same 'm'), the next chunk is scanned again along the
    whole N space but for a different 'm', and so on.

  - Besides detector data, there are also associated metadata -- e.g. the parameter
    arrays along which the scans took place. These are also Nx(...) data sets; in the
    simplest cases, they are 2D (parameter is scalar).

  - The both detector data and scan metadata are stored at specific "paths" within
    the data container. In HDF5 terminology, this is the literal HDF5 path within
    the file (e.g. "data/rayonix/detector"). We don't make any assumptions about
    the tree organisation (yes, typically it follows NX data format rules, but...).

  - A data container (i.e. a "file") can contain one or several scans. Each scan
    has a top-level subdirectory of its own.

  - Beyond data series (detector and metadata), there can also exist single scalar
    values setting the state for the entire experiment (e.g. "photon_energy", or
    "center_pixel", ...).

  - [FIXME: haven't decided yet whether top-level scalar metadata still needs to
    be embedded into a scan (i.e. is per-scan only), or can also be at the very
    top. I think ESRF has the former...]

The `source.DataSource` essentially uses the following HDF5 API elements:

  - Accessing of data using slicers and the indexing operator (`__getitem__`).

  - `__getitem__` access works both for selecting paths / subpaths, or selecting
    data containers within paths, or slicing parts of the data container itself,
    i.e. sub-access to limited parts of the images (i.e. only specific
    region-of-interests) via slicing. Examples:
    ```
      data["detector"] -> directory containing the "acme" subdirectory

      data["detector/acme"] -> directory, containing the "images" dataset
      data["detector"]["acme"] -> same as above

      data["detector/acme/images"] -> the dataset itself
      data["detector"]["acme"]["images"] -> same as above

      data["detector/acme/images"][3] -> image no. 3 from the dataset

      data["detector/acme/images"][3][60:160,60:160] -> a 100x100 pixel area centered
                                                        around pixel 110x110 of image no. 3

    ```

  - The HDF5 taxonomy (i.e. the '.attrs' dictionary

  - We're not using any HDF5 specific properties or Python attributes.

'''

'''
Experiment configuration template to use with `nx5d.xrd.data.ExperimentSetup`.
Contains goniometer and detector layout for the Pilatus detector, as well
as HDF5 addresses for the goniometer and detector angles.
'''
ExperimentTemplate = {
    "goniometerAxes": ('x+', 'y+', 'z+'),
    
    "detectorTARAxes": ("x+", None, None),
    
    "imageAxes": ("x-", "z-"),
    "imageSize": (195, 487),
    "imageCenter": (90, 245),

    # This could also be used instead of 'imageChannelSize' below.
    # It's the same physical quantity, but in degrees/channel
    # instead of relative length.
    "imageChannelSpan": None,

    "imageDistance": "@{positioners}/PilatusYOffset",
    "imageChannelSize": (0.172, 0.172), # same unit as imageDistance (mm)
    
    "sampleFaceUp": 'z+',
    "beamDirection": (0, 1, 0),
    
    "sampleNormal": (0, 0, 1),
    
    "beamEnergy": "@{positioners}/monoE",

    "goniometerAngles": {
        'theta': "@{positioners}/Theta",
        'chi':   "@{positioners}/Chi",
        'phi':   "@{positioners}/Phi" },

    "detectorTARAngles": {
        'azimuth': "@{positioners}/TwoTheta",
    }
}

class SpecH5LazyTiffNode(SpecH5LazyNodeDataset):
    '''
    Implements a NodeDataset that lazily loads a list of TIFF files,
    specified as a path format that relies on an index variable.
    '''
    def __init__(self, *args, tiffPathFmt="{frameidx}.tiff", numFrames=0, **kwargs):
        super().__init__(*args, **kwargs)
        self._tiff_path_fmt = tiffPathFmt
        self._tiff_frames_no = int(numFrames)


    def _create_data(self):
        #data = []
        #for i in range(self._tiff_frames_no):
        #    data.append( tifffile.imread(self._tiff_path_fmt.format(frameidx=i)) )
        #return np.array(data)
        return np.array([tifffile.imread(self._tiff_path_fmt.format(frameidx=i)) \
                         for i in range(self._tiff_frames_no)])


    def __getitem__(self, item):
        '''
        Apparently we can optimize this so that only a single file is
        loaded if not already initialized (...so we don't have to load
        the whole bunch for single-frame access). In that case, `item`
        is the index of the file, and all we have to do is plug it
        into the name formatter.

        OTOH the ESRF H5 library is slow af in specific instances (e.g.
        when doing itermittend slicing a 3D dataset from HDF5.) Not sure if
        this is a problem of HDF5 or of some ESRF trickery. So we might keep
        an eye on this (...in case it's something like this).
        '''
        if not self._is_initialized: 
            try:            
                idx = None
                slc = None
                if isinstance(item, int):
                    idx = item
                    slc = None
                else:
                    idx, slc = item
                    assert isinstance(idx, int)

                # Access the frame from 'idx'
                p = self._tiff_path_fmt.format(index=item)
                frame = np.array([tifffile.imread(p)])

                # ...and possibly a sub-slice ('None' is the same as [:])
                return frame[slc]

            except:
                pass

        return super().__getitem__(item)

    

class SpecTiffH5(SpecH5):
    '''
    Mocks a HDF5 file using the Spec H5 module from Silx.

    In addition the "regular" `slix.io.spech5.SpecH5` functionality, after loading the
    SPEC file, this module also injects lazy-loading nodes for the TIFF files
    correspondng to a typical KMC3 setup.
    '''

    def __init__(self, *args, instrumentName="pilatus",
                 framePathFmt='{dirname}/{instr}/S{scannr:05}/{basename}_{scannr}_{frameidx}.tif',
                 **kwargs):
        '''
        Parameters:

          - `instrument`: A name/label for the instrument. For one, this is used to compile
            the location of where to insert the data from the external TIFF files (as
            specified by `tiffInsertFmt` and `tiffLinkFmt`); for another, this is used
            to determine the location of the TIFF files on disk relative to the
            SPEC file (as specified by `scanPathFmt`).
        
          - `framePathFmt`: format for the image scan filepath.

          - `scanIndexNameFmt`: string formatting that ties the scan number (on disk)
            and scan name (in teh SpecH5 naming) together.

         All string formatters can make use of any or all of the following variables:
        
           - `{dirname}`: folder-only component of the spec file name
        
           - `{basename}`: base name the SPEC file, with the last component after a dot
             (typically ".spec" removed

           - `{scannr}`: scan number; this is typically an integer, assigned by "the
             algorithm", and typically starting with 1 for the first scan

           - `{scanidx}`: scan index; this is an integer starting with 0 and designating
             the cosecutive index value of the scan.

           - `{frameidx}`: the index (starting with 0) of the frame being processed
             (i.e. the TIFF file).

        In addition to the explicit parameters, we also use the following positional
        parameters from `args`:

           - `filename` (position 0): the path of the SPEC file to read
        
        '''
        super().__init__(*args, **kwargs)
        
        self._specBaseName = '.'.join(path.basename(self.filename).split('.')[:-1]) \
            or path.base(self.filename)
        self._specDirName = path.abspath(path.dirname(self.filename))

        for scanName in self.keys():
            
            scan = self[scanName]
            scanNr, foo = scanName.split('.')

            tiffPartialFmt = framePathFmt.format(**{
                "basename": self._specBaseName,
                "dirname": self._specDirName,
                "scannr": int(scanNr),
                "instr": instrumentName,
                "scanidx": "{scanidx}",
                "frameidx": "{frameidx}"
            })
            
            self._adopt_scan(scan, instrumentName, tiffPartialFmt)

            self._enter_cnt = 0
            self._enter_data = None


    def __enter__(self, *args, **kwargs):
        '''
        Opening / closing a SpecTiffH5 is pretty expensive. We usually want to
        employ with-guards (enter/exit) when doing that, but the original SpecH5
        does not support "nested" guards (i.e. entering/exitting multiple
        times on the same object).

        This is an extension to do just that. (Need to check if h5py.File supports
        this... the original Python open() / file object definitely does.)
        '''
        if self._enter_cnt == 0:
            self._enter_data = super().__enter__(*args, **kwargs)
        self._enter_cnt += 1
        return self._enter_data


    def __exit__(self, *args, **kwargs):
        if self._enter_cnt > 0:
            self._enter_cnt -= 1
        if self._enter_cnt == 0:
            super().__exit__(*args, **kwargs)


    def _adopt_scan(self, scanObj, instrumentName, tiffFormat):

        measRoot = scanObj['measurement']
        instrRoot = scanObj['instrument']
        posnrs = instrRoot['positioners']
        

        # Let's hope they're all the same... :-)
        numFrames = 0
        #print("Positioners:", posnrs.keys())
        for p in posnrs.keys():
            #print("Positioner:", p, "length", posnrs[p].size)
            numFrames = max(numFrames, posnrs[p].size)

        # place data inside "instrument/<name>/data"
        instrRoot.add_node(H5Group(parent=instrRoot, name=instrumentName,
                                   attrs={'NX_class': 'NXinstrument'} ))
        instrObj = instrRoot[instrumentName]

        instrObj.add_node(SpecH5LazyTiffNode(tiffPathFmt=tiffFormat,
                                             numFrames=numFrames,
                                             name="data",
                                             parent=instrObj))

        # create a link inside "measurement" that points to "instrument/<name>/data"
        link_target = instrObj.name.replace('measurement', 'instrument')
        measRoot.add_node(H5SoftLink(name=instrumentName,
                                     path=link_target + "/data",
                                     parent=measRoot))

        
def test_kmc3nx():
    
    k = Kmc3NxMocker(specFile="/var/home/specuser/tmp/nxdata/231-cw7-12083-roessle.spc/"
                     "231-cw7-12083-roessle.spec", scanNr=3)

    for f in k.dataFiles:
        assert path.isfile(f)



#!/usr/bin/python3

import pytest
from nx5d.xrd import kmc3 as kmc3

@pytest.fixture
def testfile():

    # Note that, for data storage reasons within the git project,
    # our testfile contains only TIFF data for scan "2.1".
    return "./tests/test_data/spech5/231-cw7-12083-roessle.spec"


def test_spectiff(testfile):

    h5like = kmc3.SpecTiffH5(testfile)

    assert "2.1" in h5like
    assert "44.1" in h5like ## scan is available, but no tiff data

    scan = h5like["2.1"]

    for i in [ "measurement", "instrument" ]:
        assert i in scan

    assert "pilatus" in scan["measurement"]
    assert "data" in scan["instrument/pilatus"]

    d1 = scan['measurement/pilatus']
    d2 = scan['instrument/pilatus/data']

    for i,j in zip(d1.shape, d2.shape):
        assert i == j
        assert j != 0

    # There's only one image in scan "2.1"
    assert len(d1) == 1

    assert (d1 == d2).all()
    



#!/usr/bin/python3

import xrayutilities as xu
from xarray import DataArray
import numpy as np

import logging

class LazyQMap:
    '''
    Container class for a stack of experimental data frames
    and associated metadata.

    The raw / authoritative data is stored in the `Chunk.data`
    container, which has a Python `dict()` like interface.
    Beyond that, this class also offers a number of processed
    data entries that are all lazily evaluated (i.e. only processed
    on first access):

      - `qimg`: A single 2D image of the Q-space converted data stack,
        on a rectantular grid (using an `xrayutilities` gridder, meaning
        that *all* of the original image stack is being used)

      - `qxaxis`: The X-axis of the gridder that produced `qimg`

      - `qyaxis`: The Y-axis of the gridder that produced `qimg`

    All of the computed data being *lazily* evaluated means that any
    processing that must take place on the raw (i.e. untransformed)
    data can -- and must -- take place before first access to any
    of the `q...` properties. E.g. for intensity normalization, you
    could do simething like: `chunk.data['img'] *= intensity` and
    only then proceed to accessing `chunk.qimg`.
    '''

    def __init__(self, setup=None, **data):
        '''
        Parameters:
          - `data`: A dicitonary with string keys (names) and arrays
            as the data fields.

          - `setup`: Experiment definition dictionary (see for instance
            nx5d.xrd.kmc3.ExperimentTemplate).

        '''
        
        self.data = data
        
        self.angles = {}
        if len(self.angles) == 0:
            self.angles.update(**(setup['goniometerAngles']))
            self.angles.update(**(setup['detectorTARAngles']))
            
        self.hxrd = self.__init_experiment(setup)
        self.Ang2Q = self.hxrd.Ang2Q


    def __init_experiment(self, setup):
        '''
        Initializes the Experiment part (i.e. xrayutilities HXRD object etc)
        with specified device and sample geometry. The optional parameter `roi`
        restricts angle-to-Q conversion to solely this region, if it is
        specified. This is a good way to save significant amounts of computing
        time.

        '''

        logging.debug("Expriment setup: %r" % setup)

        detAxes = [x for x in filter(lambda x: x is not None, setup['detectorTARAxes'])]

        qconv = xu.experiment.QConversion(sampleAxis=setup['goniometerAxes'],
                                          detectorAxis=detAxes,
                                          r_i=setup['beamDirection'],
                                          en=setup['beamEnergy'])

        hxrd = xu.HXRD(idir=setup['beamDirection'],
                       ndir=setup['sampleNormal'],
                       sampleor=setup['sampleFaceUp'],
                       qconv=qconv,
                       # Workaround for buggy xrayutilities: repeat the beam energy
                       en=setup['beamEnergy'])

        imageCenter = setup['imageCenter']
        if imageCenter is None:
            raise RuntimeError("Incomplete experimental definition (image center not defined)")

        assert len(imageCenter) == 2

        if imageCenter[0] <= 1 and imageCenter[1] <= 1:
            # It's a single number, all-in-one
            imgs = setup['imageSize']
            assert imgs is not None
            assert imgs[0] is not None
            assert imgs[1] is not None
            imageCenter = tuple([c*s for c,s in zip(setup['imageCenter'], imgs)])

        chSizeParm = {}
        if ('imageChannelSpan' in setup) and (setup['imageChannelSpan'] is not None):
            # channelSpan is degrees/channel, but we need to pass channels/degree to Ang2Q
            chSizeParm = {'chpdeg1': 1.0/setup['imageChannelSpan'][0],
                          'chpdeg1': 1.0/setup['imageChannelSpan'][1] }

        elif ('imageChannelSize' in setup) and (setup['imageChannelSize'] is not None):
            # Ang2Q takes one explicit distance parameter, but we're assuming that channelSize
            # is relative to the distance itself (putting the distance always at 1.0 units)
            chSizeParm = { 'pwidth1':  setup['imageChannelSize'][0],
                           'pwidth2':  setup['imageChannelSize'][1],
                           'distance': setup['imageDistance'] }

        else:
            raise RuntimeError("Experiment setup needs either the channel span or channel size")

        tarAngles = setup.get('detectorTARAngles', [0, 0, 0])

        roi = setup.get('roi', (0, setup['imageSize'][0], 0, setup['imageSize'][1]))

        hxrd.Ang2Q.init_area(detectorDir1=setup['imageAxes'][0],
                             detectorDir2=setup['imageAxes'][1],
                             cch1=imageCenter[0],
                             cch2=imageCenter[1],
                             Nch1=setup['imageSize'][0],
                             Nch2=setup['imageSize'][1],
                             tiltazimuth=0, #tarAngles[1],
                             tilt=0,        #tarAngles[0],
                             detrot=0,      #tarAngles[2],
                             roi=roi,
                             **chSizeParm)

        return hxrd    


    def __getitem__(self, label):
        return self.data[label]


    def __call__(self, *datasets, **kwargs):
        return self.area_qconv(*datasets, **kwargs)

    def area_qconv(self, *datasets, **kwargs):
        '''
        Front to the ang-to-Q conversion, currently only for area data. Parameters:
        `datasets` is either empty, or a single data label. (No multiple label support
        yet.)
        
        Accepted `kwargs`:

          - `qsize`: Tuple (w, h) of the resulting Q image, or `None` (default).
            If it is `None`, the size of the original angular image(s) is used.

          - `dims`: List with dimension names for resulting Q image.
            Defaults to `["qw", "qh"]`.

          - `_gridderDict`: If specified, this is a dictionary with extra named
            parameters to be passed on to the gridder. Note that this is not portable,
            only works as long as we're using xrayutilities under the hood.

          - `_ang2qDict`: Extra set of parameters to be passed to the data-specific
            `Ang2Q` function (typically `Ang2Q.area()` for stacks of 2D datasets).
        '''

        if len(datasets) > 1:
            raise RuntimeError("Can't transform more than one dataset at a time")

        label = next(iter(datasets  if len(datasets)>0 else self.data.keys()))
        img = self.data[label]
        
        dims = kwargs.setdefault('dims', ('qx', 'qy', 'qz'))

        if isinstance(dims, str):
            dims = (dims,)

        if len(dims) < 1 or len(dims) > 3:
            raise RuntimeError("Bad Q-axis specification: %r" % dims)
        
        if len(img.shape) != 3:
            raise RuntimeError("Don't know how to transform objects of shape %r" % (img.shape,))


        # Make sure all angles are arrays (even those that might have been passed as floats)
        ang = [a if hasattr(a, "__len__") else np.array([a]*len(img)) \
               for a in self.angles.values()]
            
        self.q = self.Ang2Q.area(*ang, **(kwargs.get('_ang2qDict', {})))

        # For transforming strings to dimension indices
        qspec = { 'x': 0, 'qx': 0,
                  'y': 1, 'qy': 1,
                  'z': 2, 'qz': 2 }

        # The list of individual Q directions/coordinates, according to user input 'dims'
        qcoord = [self.q[qspec[d]] for d in dims]

        # Default for 'qsize' is the size of the final Q-image -- make sure we account
        # for the possibly twisted order of Q-coordinates the user specified in 'dims'
        kwargs.setdefault('qsize', [img.shape[qspec[d]] for d in dims])

        # Call scheme of all the xrayutilities Gridders is pretty similar.
        Gridder = getattr(xu, "FuzzyGridder%dD" % len(dims))
        grd = Gridder(*(kwargs['qsize']))
        grd( *qcoord, img, **(kwargs.get('_gridderDict', {})) )

        # ...the tricky part is creating the DataArrays. Specifically,
        # retrieving the q coordinates from the gridder. They are in `grd`
        # attributes called 'xaxis', 'yaxis', ... according to dimension.
        # We always use qx/qy/qz for dimension keys.
        coords = {}
        for i,d in enumerate(dims):
            axname = d if len(d)>1 else 'q%s' % d
            axvals = getattr(grd, "%saxis" % ('x', 'y', 'z')[i])
            coords[axname] = axvals

        return DataArray(data=grd.data, coords=coords)

    
## Example for a class that does more than LazyQMap (namely accept angle offsets
## in its constructor), but still uses LazyQMap under the hood.
class OffsetQMap(LazyQMap):
    def __init__(self, offsets=None, **kwargs):
        super().__init__(**kwargs)
        if offsets is not None:
            for k,v in offsets.items():
                self.angles[k] += v



#!/usr/bin/python3

import numpy as np

'''
Helps working with N-dimensional region of interests (ROIs) in data.
'''

class Roi:
    '''
    Roi represents a region of interest in N dimensions.
    It can be initialized in a variety of ways:
      - By explicitly specifying start and end points
      - By specifying a start point and the extent
      - By specifying a center and an extent
    Once set up, slicers can be generated that select the
    corresponding data points from each dimension of an
    N-dimensional array.
    '''
    
    def __init__(self, start=None, size=None, center=None, end=None,
                 itype=None, dims=None):
        '''
        Initializes the ROI. Each of the parameters can be iterables
        (arrays, lists, tuples) or scalars. All have to have the
        same dimension and type.
        
        The ROI coordinates can be integers (default) or other types,
        e.g. floating point numbers. The slicers that are returned
        by the `Roi.s` property will have a matching type.
        
        Parameters:
        
          - `start`: The start point (or corner) of the ROI.
          
          - `center`: As an alternative to `start`, the center point
            of an ROI can be specified with this parameter.
          
          - `end`: This represents end point (or corner) of the ROI,
            if the start point is known.
          
          - `size`: This is the extent of the ROI. If `start` was
            specified, the ROI extends to `start+size`. Size can
            be negative, in which case the end point will be before
            start. If on the other hand `center` was specified,
            then `size` is just the total extent of the ROI. Then
            the sign of `size` is ignored, only its magnitude in
            every direction is used.
            
          - `itype`: The data type (`numpy` lingo `dtype`) of the
            index values. The outgoing slicers are going to be
            of the same value. If it is `None` (default), the
            type of the input data is matched.
            
          - `dims`: If specified, it's the name of the
            dimensions. If available, the `Roi.ds` property will
            return an indexing dictionary using the dimension names
            as keys, suitable for slicing an `xarray`.
        '''
        
        ## Internally we only work with "bottom-left" and "top-right"
        ## values, i.e. the lowest and the highest of each dimension ¯\_(ツ)_/¯
        
        self.__is_scalar = not hasattr(start or center, "__iter__")
                    
        if start is not None:
            start = np.array((start,) if self.__is_scalar else start)
                
        if size is not None:
            size = np.array((size,) if self.__is_scalar else size)
                
        if center is not None:
            center = np.array((center,) if self.__is_scalar else center)
                
        if end is not None:
            end = np.array((end,) if self.__is_scalar else end)
    
        # Need to recyle the data type of the input
        dtype = itype or np.dtype(type(start[0] if start is not None else center[0]))
        
        corner1 = np.array(start) if start is not None \
                    else (np.array(center)-np.abs(np.array(size))/2).astype(dtype)
        
        corner2 = np.array(end, dtype=dtype) if end is not None \
                    else corner1 + np.array(size)
        
        #print(corner1, corner2)
        
        self.bl = np.array([np.min(np.array([a,b])) for a,b in zip(corner1, corner2)])
        self.tr = np.array([np.max(np.array([a,b])) for a,b in zip(corner1, corner2)])
        
        #print (self.bl, self.tr)
        
        self.dims = dims
        
    @property
    def s(self):
        '''
        Returns a tuple if slice() objects, one for each dimension
        '''
        if self.__is_scalar:
            return slice(self.bl[0],self.tr[0])
        else:
            return tuple([slice(s,e) for s,e in zip(self.bl, self.tr)])
        
    @property
    def ds(self):
        '''
        Returns a dictionary with dimension names as keys and
        slicer objects (similarly to `Roi.s`) as values.
        '''
        return { k:v for k,v in zip(self.dims, self.s) }



#!/usr/bin/python3

'''
Geometry and data set definitions for the ESRF ID09 XRD beamline.
Need these for the `nx5d.xrd.data.ScanReader` object.
'''

'''
See `nx5d.xrd.data.ScanReader.__init_experiment()` for a more detailed
documentation of the device setup data.
'''

ExperimentTemplate = {
    "goniometerAxes": ('y+',),
    "detectorTARAxes": (None, None, None),
    "imageAxes": ("z-", "y-"),
    "imageSize": (4096, 4096),
    "imageCenter": (0.5, 0.5),

    # pixel size is typically present in the HDF5 data file, distance is 0.2 m.
    "imageChannelSize": None,

    "sampleFaceUp": 'z+',
    "sampleNormal": (0, 0, 1),
    "beamDirection": (-1, 0, 0),

    "goniometerAngles": [ "hphi" ],
    "detectorTARAngles": []
}

#DataLabels = {
        #"img":  lambda src: "measurement/rayonix",
        #"hphi": lambda src: "instrument/positioners/hphi",
        #"lxt":  lambda src: "instrument/positioners/lxt_ps",
        #"norm": lambda src: "measurement/sbcurr"
#}
